# Copyright (C) 2012-2014 Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import sys

from lmi.shell.compat import *

from lmi.shell.LMIBaseObject import LMIWrapperBaseObject
from lmi.shell.LMIFormatter import LMIClassFormatter
from lmi.shell.LMIConstantValues import LMIConstantValuesParamProp
from lmi.shell.LMIObjectFactory import LMIObjectFactory

from lmi.shell.LMIExceptions import LMIUnknownPropertyError

from lmi.shell.LMIDecorators import lmi_class_fetch_lazy
from lmi.shell.LMIDecorators import lmi_process_cim_exceptions
from lmi.shell.LMIDecorators import lmi_return_val_if_fail
from lmi.shell.LMIDecorators import lmi_return_if_fail

from lmi.shell.LMIShellLogger import lmi_get_logger

from lmi.shell.LMIUtil import lmi_cast_to_cim
from lmi.shell.LMIUtil import lmi_get_use_exceptions
from lmi.shell.LMIUtil import lmi_set_use_exceptions
from lmi.shell.LMIUtil import lmi_raise_or_dump_exception
from lmi.shell.LMIUtil import lmi_transform_to_lmi


class LMIClass(LMIWrapperBaseObject):
    """
    LMI wrapper class representing :py:class:`wbem.CIMClass`.

    :param LMIConnection conn: connection object
    :param LMINamespace namespace: namespace object
    :param string classname: CIM class name
    """
    def __init__(self, conn, namespace, classname):
        # We use __dict__ to avoid recursion potentially caused by
        # combo __setattr__ and __getattr__
        self.__dict__["_namespace"] = namespace
        self.__dict__["_cim_class"] = None
        self.__dict__["_cim_class_full_fetch"] = False
        self.__dict__["_cim_classname"] = classname
        self.__dict__["_valuemap_properties_list"] = []
        super(LMIClass, self).__init__(conn)

    def __repr__(self):
        """
        :returns: a pretty string for the object
        """
        return "%s(classname='%s', ...)" % (
            self.__class__.__name__, self.classname)

    def __getattr__(self, name):
        """
        Returns either a class member, or a constant value.

        Simplifies the code and constant value can be retrieved by
        :samp:`object.constant_value`.

        :param string name: object member or constant values member
        :returns: class member of :py:class:`.LMIConstantValuesParamProp`
            object
        """
        if not self._conn.is_wsman() and name.endswith("Values"):
            if not self.is_fetched(True):
                self.fetch(True)
            property_name = name[:-6]
            return LMIConstantValuesParamProp(
                self._cim_class.properties[property_name])
        raise AttributeError(name)

    @lmi_class_fetch_lazy()
    @lmi_return_val_if_fail(lambda obj: obj._namespace, None)
    def create_instance(self, properties=None, qualifiers=None,
                        property_list=None):
        """
        Creates a new :py:class:`wbem.CIMInstance` at the server side and
        returns :py:class:`.LMIReturnValue` object containing
        :py:class:`.LMIInstance` as a result.

        :param dictionary properties: initial properties with corresponding
            values
        :param dictionary qualifiers: initial qualifiers
        :param list property_list: list of properties, which should be present
            in :py:class:`.LMIInstance` object

        **Usage:** See :ref:`class_create_instance`.
        """
        if self._conn.is_wsman():
            # WSMAN client doesn't support CreateInstance()
            lmi_get_logger().info("WSMAN client doesn't support CreateInstance()")
            return None

        # No need to copy dictionaries to avoid the variable mix-up, the
        # copying is done in client.create_instance(), we just pass
        # what we get.
        properties = properties if properties is not None else {}
        qualifiers = qualifiers if qualifiers is not None else {}
        self_properties = self._cim_class.properties
        for key, value in properties.iteritems():
            if key not in self._cim_class.properties:
                errorstr = "No such instance property '%s'" % key
                lmi_raise_or_dump_exception(LMIUnknownPropertyError(errorstr))
                return None
            if isinstance(value, LMIObjectFactory().LMIInstanceName):
                value = value.wrapped_object
            t = self._cim_class.properties[key].type
            properties[key] = lmi_cast_to_cim(t, value)
        instance, rparams, errorstr = self._conn.client.create_instance(
            self.classname, self.namespace, self._conn.hostname,
            properties, qualifiers, property_list)
        if not instance:
            return None
        return lmi_transform_to_lmi(self._conn, instance)

    @lmi_class_fetch_lazy(True)
    @lmi_return_if_fail(lambda obj: obj._namespace)
    def doc(self):
        """
        Prints out pretty verbose message with documentation for the class. If
        the LMIShell is run in a interactive mode, the output will be
        redirected to a pager set by environment variable :envvar:`PAGER`. If
        there is not :envvar:`PAGER` set, less or more will be used as a
        fall-back.
        """
        if self._conn.is_wsman():
            # WSMAN doesn't support reflextion, we can't call GetClass().
            lmi_get_logger().info("WSMAN client doesn't support GetClass()")
            return

        LMIClassFormatter(self._cim_class).fancy_format(
            self._conn.client.interactive)

    def is_fetched(self, full_fetch=False):
        """
        Returns True, if :py:class:`wbem.CIMClass` has been fetched.

        :param bool full_fetch: defines, if qualifiers are also included
        """
        if self._conn.is_wsman():
            # WSMAN doesn't support GetClass(). We return always False.
            return False

        return bool(self._cim_class) and \
            (not full_fetch or self._cim_class_full_fetch)

    @lmi_process_cim_exceptions(
        lambda x, y: None,
        lambda x, y: None)
    def fetch(self, full_fetch=False):
        """
        Manually fetches a wrapped :py:class:`wbem.CIMClass` object.

        :param bool full_fetch: True, if :py:class:`wbem.CIMClass` should
            include qualifiers and class origin. Default value is False.
        :raises: :py:exc:`.CIMError`, :py:exc:`.ConnectionError`

        **Usage:** See :ref:`class_fetching_a_class`.
        """
        if self._conn.is_wsman():
            # WSMAN doesn't support GetClass(). We do not try to fetch the
            # class.
            return

        if self.is_fetched(full_fetch) or not self._namespace:
            # We already have CIMClass (at certain level of detail), or we do
            # not know the namespace, from which the class should be fetched.
            return

        full_cim_class, _, _ = self._conn.client.get_class(
            self._cim_classname,
            self._namespace.name,
            full_fetch=full_fetch,
            LocalOnly=False)

        self._cim_class = full_cim_class

        if full_fetch and self._cim_class:
            self._cim_class_full_fetch = full_fetch
            # Store the constant values as a list. This can consume some time,
            # if computed on demand.
            self._valuemap_properties_list = [
                k for k, v in self._cim_class.properties.iteritems()
                if "ValueMap" in v.qualifiers and "Values" in v.qualifiers]

    # NOTE: usage with Key=something, Value=something is deprecated
    # NOTE: inst_filter is either None or dict
    @lmi_return_val_if_fail(lambda obj: obj._namespace, [])
    def instance_names(self, inst_filter=None, **kwargs):
        """
        Returns a LMIReturnValue containing a list of LMIInstanceNames.

        :param dictionary inst_filter: filter values. The key corresponds to
            the primary key of the :py:class:`wbem.CIMInstanceName`; value
            contains the filtering value
        :param dictionary kwargs: deprecated keyword arguments

            * **Key** or **key** (*string*) -- filtering key, see above
            * **Value** or **value** -- filtering value, see above

        :returns: :py:class:`.LMIReturnValue` object with ``rval`` set to a
            list of :py:class:`.LMIInstanceName` objects

        **Usage:** See :ref:`class_get_instance_names` and
        :ref:`class_instance_filtering`.
        """
        inst_name_list, _, errorstr = self._conn.client.get_instance_names(
            self._cim_classname, self._namespace.name, inst_filter, **kwargs)
        if not inst_name_list:
            return []
        return lmi_transform_to_lmi(self._conn, inst_name_list)

    def new_instance_name(self, keybindings):
        """
        Create new :py:class:`.LMIInstanceName` object by passing all the
            keys/values of the object.

        :param dictionary keybindings: primary keys of instance name with
            corresponding values
        :returns: new :py:class:`.LMIInstanceName` object

        **Usage:** See :ref:`class_new_instance_name`.
        """
        kbs = wbem.NocaseDict()
        for key, value in keybindings.iteritems():
            if isinstance(value, LMIObjectFactory().LMIInstanceName):
                value = value.wrapped_object
            elif isinstance(value, str):
                # Convert strings to unicode
                value = unicode(value, "utf-8")
            kbs[key] = value
        cim_inst_name = wbem.CIMInstanceName(
            self.classname, kbs, namespace=self.namespace)
        return lmi_transform_to_lmi(self._conn, cim_inst_name)

    # NOTE: usage with Key=something, Value=something is deprecated
    # NOTE: inst_filter is either None or dict
    def first_instance_name(self, inst_filter=None, **kwargs):
        """
        Returns the first :py:class:`.LMIInstanceName` of the corresponding
        class.

        :param dictionary inst_filter: filter values, where the key corresponds
            to the primary key of :py:class:`wbem.CIMInstanceName`; value
            contains the filtering value
        :param dictionary kwargs: deprecated keyword arguments

            * **Key** or **key** (*string*) -- filtering key, see above
            * **Value** or **value** -- filtering value, see above

        :returns: first :py:class:`.LMIInstanceName` object

        **Usage:** See :ref:`class_get_instance_names` and
        :ref:`class_instance_filtering`.
        """
        inst_name_list, _, errorstr = self._conn.client.get_instance_names(
            self._cim_classname, self._namespace.name, inst_filter, limit=1,
            **kwargs)
        if not inst_name_list:
            return None
        return lmi_transform_to_lmi(self._conn, inst_name_list[0])

    # NOTE: usage with Key=something, Value=something is deprecated
    # NOTE: inst_filter is either None or dict
    @lmi_return_val_if_fail(lambda obj: obj._namespace, [])
    def instances(self, inst_filter=None, client_filtering=False, **kwargs):
        """
        Returns a list of objects of :py:class:`.LMIInstance`.

        :param dictionary inst_filter: filter values, where the key corresponds
            to the key of :py:class:`wbem.CIMInstance`; value contains the
            filtering value
        :param bool client_filtering: if True, client-side filtering will be
            performed, otherwise the filtering will be done by a CIMOM. Default
            value is False.
        :param dictionary kwargs: deprecated keyword arguments

            * **Key** or **key** (*string*) -- filtering key, see above
            * **Value** or **value** -- filtering value, see above

        :returns: list of :py:class:`.LMIInstance` objects

        **Usage:** See :ref:`class_get_instances` and
        :ref:`class_instance_filtering`.
        """
        instance_list, _, errorstr = self._conn.client.get_instances(
            self._cim_classname, self._namespace.name, inst_filter,
            client_filtering, **kwargs)
        if not instance_list:
            return []
        return lmi_transform_to_lmi(self._conn, instance_list)

    # NOTE: usage with Key=something, Value=something is deprecated
    # NOTE: inst_filter is either None or dict
    def first_instance(self, inst_filter=None, client_filtering=False,
                       **kwargs):
        """
        Returns the first :py:class:`.LMIInstance` of the corresponding class.

        :param dictionary inst_filter: filter values, where the key corresponds
            to the key of :py:class:`wbem.CIMInstance`; value contains the
            filtering value.
        :param bool client_filtering: if True, client-side filtering will be
            performed, otherwise the filtering will be done by a CIMOM. Default
            value is False.
        :param dictionary kwargs: deprecated keyword arguments

            * **Key** or **key** -- filtering key, see above
            * **Value** or **value** -- filtering value, see above

        :returns: first :py:class:`.LMIInstance` object

        **Usage:** See :ref:`class_get_instances` and
        :ref:`class_instance_filtering`.
        """
        filter_value = "value" in [k.lower() for k in kwargs.keys()]
        if inst_filter or filter_value:
            instance_list, _, errorstr = self._conn.client.get_instances(
                self._cim_classname, self._namespace.name, inst_filter,
                client_filtering, limit=1, **kwargs)
            if not instance_list:
                return None
            instance = instance_list[0]
            return lmi_transform_to_lmi(self._conn, instance)
        inst_name = self.first_instance_name(inst_filter, **kwargs)
        return inst_name.to_instance() if inst_name else None

    @lmi_class_fetch_lazy(True)
    @lmi_return_val_if_fail(lambda obj: obj._namespace, [])
    def valuemap_properties(self):
        """
        :returns: list of strings of the constant names

        **Usage:** :ref:`class_get_valuemap_properties`.
        """
        return self._valuemap_properties_list

    @lmi_class_fetch_lazy(True)
    @lmi_return_if_fail(lambda obj: obj._namespace)
    def print_valuemap_properties(self):
        """
        Prints out the list of string of constant names.

        **Usage:** :ref:`class_get_valuemap_properties`.
        """
        for i in self._valuemap_properties_list:
            sys.stdout.write("%s\n" % i)

    @lmi_class_fetch_lazy()
    @lmi_return_val_if_fail(lambda obj: obj._namespace, [])
    def properties(self):
        """
        :returns: list of strings of the :py:class:`wbem.CIMClass` properties

        **Usage:** See :ref:`class_properties`.
        """
        if self._conn.is_wsman():
            # WSMAN does not support GetClass().
            return []

        return self._cim_class.properties.keys()

    @lmi_class_fetch_lazy()
    @lmi_return_if_fail(lambda obj: obj._namespace)
    def print_properties(self):
        """
        Prints out the list of :py:class:`wbem.CIMClass` properties.

        **Usage:** See :ref:`class_properties`.
        """
        for prop in self.methods():
            sys.stdout.write("%s\n" % prop)

    @lmi_class_fetch_lazy()
    @lmi_return_val_if_fail(lambda obj: obj._namespace, [])
    def methods(self):
        """
        :returns: list of strings of :py:class:`wbem.CIMClass` methods.

        **Usage:** See :ref:`class_methods`.
        **Note:** When caching is turned off, this method may consume some
        time.
        """
        if self._conn.is_wsman():
            # WSMAN does not support GetClass().
            return []

        methods_lst = self._cim_class.methods.keys()
        class_names, _, _ = self._conn.client.get_class_names(
            self.namespace, DeepInheritance=True)

        if class_names is not None:
            # Iterate through a list of all CIM methods and append any method
            # which can be called in the synchronous way. One type of such
            # method is the method, which has its counterpart class
            # __MethodParameters_<Method>.  The other one is the method, which
            # defines a Job return paremeter.
            for name, method in self._cim_class.methods.iteritems():
                if "__MethodParameters_" + name in class_names or \
                        "Job" in method.parameters:
                    methods_lst.append("Sync" + name)

        return methods_lst

    @lmi_class_fetch_lazy(True)
    @lmi_return_if_fail(lambda obj: obj._namespace)
    def print_methods(self):
        """
        Prints out the list of :py:class:`wbem.CIMClass` methods.

        **Usage:** See :ref:`class_methods`.
        """
        for name, method in self._cim_class.methods.iteritems():
            in_params = [
                "%s %s" % (param.type, pname)
                for pname, param in method.parameters.iteritems()
                if "Out" not in param.qualifiers or \
                    not param.qualifiers["Out"]]
            out_params = [
                "%s:%s" % (pname, param.type)
                for pname, param in method.parameters.iteritems()
                if "Out" in param.qualifiers and \
                    param.qualifiers["Out"]]
            sys.stdout.write(
                "LMIReturnValue(%s rval, rparams={%s}) %s(%s)\n" % (
                    method.return_type, ", ".join(out_params),
                    name, ", ".join(in_params)))

    @property
    def classname(self):
        """
        :returns: class name
        :rtype: string
        """
        return self._cim_classname

    @property
    @lmi_return_val_if_fail(lambda obj: obj._namespace, "Unknown")
    def namespace(self):
        """
        :returns: namespace name
        :rtype: string
        """
        return self._namespace.name

    @property
    @lmi_class_fetch_lazy()
    def wrapped_object(self):
        """
        :returns: wrapped :py:class:`wbem.CIMClass` object
        """
        return self._cim_class



# Copyright (C) 2012-2014 Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import os
import sys
import atexit
import getpass
import readline
import urlparse

import M2Crypto.SSL
import M2Crypto.SSL.Checker
import M2Crypto.X509

from lmi.shell.compat import *

from lmi.shell.LMICIMXMLClient import LMICIMXMLClient
from lmi.shell.LMIWSMANClient import LMIWSMANClient
from lmi.shell.LMIShellClient import LMIShellClient
from lmi.shell.LMINamespace import LMINamespaceRoot
from lmi.shell.LMIReturnValue import LMIReturnValue
from lmi.shell.LMISubscription import LMISubscription
from lmi.shell.LMIObjectFactory import LMIObjectFactory

from lmi.shell.LMIExceptions import CIMError
from lmi.shell.LMIExceptions import ConnectionError
from lmi.shell.LMIExceptions import LMIIndicationError
from lmi.shell.LMIExceptions import LMINamespaceNotFound

from lmi.shell.LMIShellLogger import lmi_get_logger

from lmi.shell.LMIUtil import lmi_raise_or_dump_exception
from lmi.shell.LMIUtil import lmi_is_localhost

logger = lmi_get_logger()


def __lmi_raw_input(prompt, use_echo=True):
    """
    Reads a string from the standard input.

    :param string prompt: input prompt
    :param bool use_echo: whether to echo the input on the command line
    :returns: input string
    :raises: :py:exc:`EOFError`
    """
    if not sys.stderr.isatty() and not sys.stdout.isatty():
        logger.warn(
            "Both stdout and stderr are detached from terminal, "
            "using stdout for prompt")
    elif not sys.stdout.isatty() and sys.stderr.isatty():
        # read the input with prompt printed to stderr
        stream = sys.stderr
    else:
        # read the input with prompt printed to stdout
        stream = sys.stdout

    # Define input function
    if use_echo:
        def get_input(prompt):
            return raw_input(prompt)
    else:
        def get_input(prompt):
            return getpass.getpass(prompt, stream)

    try:
        result = get_input(prompt)
    except EOFError, e:
        stream.write("\n")
        return None
    except KeyboardInterrupt, e:
        raise e

    if result and use_echo:
        cur_hist_len = readline.get_current_history_length()
        if cur_hist_len > 1:
            readline.remove_history_item(cur_hist_len - 1)

    return result


def connect(uri, username="", password="", interactive=False, use_cache=True,
            key_file=None, cert_file=None, verify_server_cert=True,
            prompt_prefix=""):
    """
    Creates a connection object with provided URI and credentials.

    :param string uri: URI of the CIMOM
    :param string username: account, under which, the CIM calls will be
        performed
    :param string password: user's password

    :param bool interactive: flag indicating, if the LMIShell client is running
        in the interactive mode; default value is False.
    :param bool use_cache: flag indicating, if the LMIShell client should use
        cache for :py:class:`wbem.CIMClass` objects. This saves lot's of
        communication, if there are :func:`EnumerateInstances` and
        :func:`EnumerateClasses` intrinsic methods often issued. Default value
        is True.
    :param string key_file: path to x509 key file; default value is None
    :param string cert_file: path to x509 cert file; default value is None
    :param bool verify_server_cert: flag indicating, whether a server side
        certificate needs to be verified, if SSL used; default value is True.
    :param string prompt_prefix: username and password prompt prefix in case
        the user is asked for credentials. Default value is empty string.
    :returns: :py:class:`.LMIConnection` object or None, if LMIShell does not
        use exceptions
    :raises: :py:exc:`.ConnectionError`

    **NOTE:** If interactive is set to True, LMIShell will:

    * prompt for username and password, if missing and connection via
      Unix socket can not be established.
    * use pager for the output of: :py:meth:`.LMIInstance.doc`,
      :py:meth:`.LMIClass.doc`, :py:meth:`.LMIInstance.tomof` and
      :py:meth:`.LMIMethod.tomof`

    **Usage:** :ref:`startup_connection`.
    """
    connection = None

    if lmi_is_localhost(uri) and \
            not username and not password and \
            not cert_file and not key_file:
        try:
            # Try to get socket's permissions. If we can talk to TOG-Pegasus,
            # stat() will return stat buf; OSError will be raised otherwise.
            os.stat("/var/run/tog-pegasus/cimxml.socket")

            connection = LMIConnection(
                uri, interactive=interactive, use_cache=use_cache)

            # Verify the connection
            rval, _, errorstr = connection.connect()
            if not rval:
                logger.warning("Can't connect via Unix socket")
                del connection
            else:
                logger.info("Connected via Unix socket")
                return connection
        except OSError, e:
            # It is fine not to do anything here. User will be prompted for
            # necessary credentials, if necessary.
            pass

    if interactive and not key_file and not cert_file:
        try:
            if not username:
                username = __lmi_raw_input(
                    prompt_prefix + "username: ", True)
            if not password:
                password = __lmi_raw_input(
                    prompt_prefix + "password: ", False)
        except KeyboardInterrupt, e:
            sys.stdout.write("\n")
            return None

    connection = LMIConnection(
        uri, username, password, interactive=interactive,
        use_cache=use_cache, key_file=key_file, cert_file=cert_file,
        verify_server_cert=verify_server_cert)

    # Verify the connection
    rval, _, errorstr = connection.connect()

    if not rval:
        logger.error(errorstr)
        return None
    else:
        logger.info("Connected to %s", uri)

    return connection


class LMIConnection(object):
    """
    Class representing a connection object. Each desired connection to separate
    CIMOM should have its own connection object created. This class provides an
    entry point to the namespace/classes/instances/methods hierarchy present in
    the LMIShell.

    :param string uri: URI of the CIMOM
    :param string username: account, under which, the CIM calls will be
        performed
    :param string password: user's password
    :param bool interactive: flag indicating, if the LMIShell client is running
        in the interactive mode; default value is False.
    :param bool use_cache: flag indicating, if the LMIShell client should use
        cache for CIMClass objects. This saves lot's of communication, if there
        are :func:`EnumerateInstances` and :func:`EnumerateClasses` intrinsic
        methods often issued. Default value is True.
    :param string key_file: path to x509 key file; default value is None
    :param string cert_file: path to x509 cert file; default value is None
    :param bool verify_server_cert: flag indicating, whether a server side
        certificate needs to be verified, if SSL used; default value is True

    **NOTE:** If interactive is set to True, LMIShell will:

    * prompt for username and password, if missing and connection via
      Unix socket can not be established.
    * use pager for the output of: :py:meth:`.LMIInstance.doc`,
      :py:meth:`.LMIClass.doc`, :py:meth:`.LMIInstance.tomof` and
      :py:meth:`.LMIMethod.tomof`
    """
    def __init__(self, uri, username="", password="", interactive=False,
                 use_cache=True, key_file=None, cert_file=None,
                 verify_server_cert=True):
        parse_res = urlparse.urlparse(uri)
        if parse_res.path != "/wsman":
            # We talk to CIMOM via CIM-XML
            self._client = LMIShellClient(
                uri, username, password, interactive=interactive,
                use_cache=use_cache, key_file=key_file,
                cert_file=cert_file, verify_server_cert=verify_server_cert)
        else:
            # We talk to CIMOM via WSMAN
            self._client = LMIWSMANClient(
                uri, username, password, interactive=interactive,
                key_file=key_file, cert_file=cert_file,
                verify_server_cert=verify_server_cert)

        self._indications = {}

        # Register LMIConnection.unsubscribe_all_indications() to be called at
        # LMIShell's exit.
        atexit.register(lambda: self.unsubscribe_all_indications())

    def __del__(self):
        """
        Disconnects and frees :py:class:`.LMICIMXMLClient` object.
        """
        self._client.disconnect()
        del self._client

    def __repr__(self):
        """
        :returns: pretty string for the object.
        """
        return "%s(URI='%s', user='%s'...)" % (
            self.__class__.__name__, self._client.uri, self._client.username)

    @property
    def client(self):
        """
        :returns: CIMOM client
        :rtype: :py:class:`.LMICIMXMLClient` or :py:class:`.LMIWSMANClient`
        """
        return self._client

    @property
    def uri(self):
        """
        :returns: URI of the CIMOM
        :rtype: string
        """
        return self._client.uri

    @property
    def hostname(self):
        """
        :returns: hostname of CIMOM
        :rtype: string
        """
        return self._client.hostname

    @property
    def namespaces(self):
        """
        :returns: list of all available namespaces

        **Usage:** :ref:`namespaces_available_namespaces`.
        """
        return ["root"]

    @property
    def root(self):
        """
        :returns: :py:class:`.LMINamespaceRoot` object for *root* namespace
        """
        return LMINamespaceRoot(self)

    @property
    def timeout(self):
        """
        :returns: CIMOM connection timeout for a transaction (milliseconds)
        :rtype: int
        """
        return self._client._cliconn.timeout

    @timeout.setter
    def timeout(self, timeout):
        """
        Sets CIMOM connection timeout for a transaction (milliseconds).

        :param inst timeout: timeout in milliseconds
        """
        self._client._cliconn.timeout = timeout

    def print_namespaces(self):
        """
        Prints out all available namespaces.
        """
        sys.stdout.write("root\n")

    def get_namespace(self, namespace):
        """
        :param string namespace: namespace path (eg. `root/cimv2`)
        :returns: :py:class:`.LMINamespace` object
        :raises: :py:exc:`.LMINamespaceNotFound`
        """
        def get_namespace_priv(namespace, namespace_path):
            if not namespace_path:
                return namespace
            ns = namespace_path.pop(0)
            return get_namespace_priv(getattr(namespace, ns), namespace_path)

        namespace_path = namespace.split("/")
        ns = namespace_path.pop(0)
        if ns not in self.namespaces:
            raise LMINamespaceNotFound(ns)
        return get_namespace_priv(getattr(self, ns), namespace_path)

    def clear_cache(self):
        """
        Clears the cache.
        """
        if hasattr(self._client, "cache"):
            self._client.cache.clear()

    def use_cache(self, active=True):
        """
        Sets a bool flag, which defines, if the LMIShell should use a cache.

        :param bool active: whether the LMIShell's cache should be used
        """
        if hasattr(self._client, "cache"):
            self._client.cache.active = active

    def connect(self):
        """
        Connects to CIMOM and verifies credentials by performing a "dummy"
        request.

        :returns: :py:class:`.LMIReturnValue` object with rval set to True, if
            the user was properly authenticated; False otherwise. In case of
            any error, rval is set to False and errorstr contains appropriate
            error string.
        :rtype: :py:class:`.LMIReturnValue`
        """
        # Connect to CIMOM. We may fail before the actual "dummy" request is
        # sent. This only applies for LMIWBEM.
        rval, _, errorstr = self._client.connect()
        if not rval:
            return LMIReturnValue(rval=False, errorstr=errorstr)

        # Perform "dummy" request to check the connection and credentials.
        return self._client.dummy()

    def disconnect(self):
        """
        Disconnects from CIMOM.
        """
        self._client.disconnect()

    def is_wsman(self):
        """
        Returns True, if the connection is made with WSMAN CIMOM; False
        otherwise.
        """
        return isinstance(self._client, LMIWSMANClient)

    def subscribe_indication(self, **kwargs):
        """
        Subscribes to an indication. Indication is formed by 3 objects, where 2
        of them (filter and handler) can be provided, if the LMIShell should
        not create those 2 by itself.

        **NOTE:** Currently the call registers :py:mod:`atexit` hook, which
        auto-deletes all subscribed indications by the LMIShell.

        :param dictionary kwargs: parameters for the indication subscription

            * **Filter** (*LMIInstance*) -- if provided, the
              :py:class:`.LMIInstance` object will be used instead of creating
              a new one;
              **optional**
            * **Handler** (*LMIInstance*) -- if provided, the
              :py:class:`.LMIInstance` object will be used instead of creating
              a new one; **optional**
            * **Query** (*string*) -- string containing a query for the
              indications filtering
            * **QueryLanguage** (*string*) -- query language; eg. *WQL*, or
              *DMTF:CQL*.  This parameter is optional, default value is
              *DMTF:CQL*.
            * **Name** (*string*) -- indication name
            * **CreationNamespace** (*string*) -- creation namespace. This
              parameter is optional, default value is *root/interop*.
            * **SubscriptionCreationClassName** (*string*) -- subscription
              object class name. This parameter is optional, default value is
              *CIM_IndicationSubscription*.
            * **Permanent** (*bool*) -- whether to preserve the created
              subscription on LMIShell's quit. Default value is False.
            * **FilterCreationClassName** (*string*) -- creation class name of
              the filter object. This parameter is options, default value is
              *CIM_IndicationFilter*.
            * **FilterSystemCreationClassName** (*string*) -- system creation
              class name of the filter object. This parameter is optional,
              default value is *CIM_ComputerSystem*.
            * **FilterSourceNamespace** (*string*) -- local namespace where the
              indications originate. This parameter is optional, default value
              is *root/cimv2*.
            * **HandlerCreationClassName** (*string*) -- creation class name of
              the handler object. This parameter is optional, default value is
              *CIM_IndicationHandlerCIMXML*.
            * **HandlerSystemCreationClassName** (*string*) -- system creation
              name of the handler object. This parameter is optional, default
              value is *CIM_ComputerSystem*.
            * **Destination** (*string*) -- destination URI, where the
              indications should be delivered

        :returns: :py:class:`.LMIReturnValue` object with ``rval`` set to True,
            if indication was subscribed; False otherwise. If a error occurs,
            ``errorstr`` is set to appropriate error string.
        """
        if self.is_wsman():
            # TODO: support WSMAN events
            return LMIReturnValue(
                rval=False,
                errorstr="Indication subscription not supported")
        try:
            indication_namespace = kwargs.get(
                "CreationNamespace", "root/interop")
            cim_filter_provided = "Filter" in kwargs
            if cim_filter_provided:
                filt = kwargs["Filter"]
                cim_filter = None
                if isinstance(filt, LMIObjectFactory().LMIInstance):
                    cim_filter = filt._cim_instance
                elif isinstance(filt, wbem.CIMInstance):
                    cim_filter = filt
                else:
                    errorstr = "Filter argument accepts instances of " \
                        "CIMInstance or LMIInstance"
                    lmi_raise_or_dump_exception(LMIIndicationError(errorstr))
                    return LMIReturnValue(rval=False, errorstr=errorstr)
            else:
                cim_filter_props = {
                    "CreationClassName": kwargs.get(
                        "FilterCreationClassName",
                        "CIM_IndicationFilter"),
                    "SystemCreationClassName": kwargs.get(
                        "FilterSystemCreationClassName",
                        "CIM_ComputerSystem"),
                    "SourceNamespace": kwargs.get(
                        "FilterSourceNamespace",
                        "root/cimv2"),
                    "SystemName": self._client.uri,
                    "Query": kwargs["Query"],
                    "QueryLanguage": kwargs.get(
                        "QueryLanguage",
                        LMICIMXMLClient.QUERY_LANG_CQL),
                    "Name": kwargs["Name"] + "-filter"
                }
                cim_filter, _, errorstr = self._client.create_instance(
                    cim_filter_props["CreationClassName"],
                    indication_namespace,
                    self.hostname,
                    cim_filter_props
                )
                if not cim_filter:
                    lmi_raise_or_dump_exception(LMIIndicationError(errorstr))
                    return LMIReturnValue(rval=False, errorstr=errorstr)
            cim_handler_provided = "Handler" in kwargs
            if cim_handler_provided:
                cim_handler = kwargs["Handler"]._cim_instance
            else:
                cim_handler_props = {
                    "CreationClassName": kwargs.get(
                        "HandlerCreationClassName",
                        "CIM_IndicationHandlerCIMXML"),
                    "SystemCreationClassName": kwargs.get(
                        "HandlerSystemCreationClassName",
                        "CIM_ComputerSystem"),
                    "SystemName": self._client.uri,
                    "Destination": "%s/CIMListener/%s" % (
                        kwargs["Destination"],
                        kwargs["Name"]),
                    "Name": kwargs["Name"] + "-handler"
                }
                cim_handler, _, errorstr = self._client.create_instance(
                    cim_handler_props["CreationClassName"],
                    indication_namespace,
                    self.hostname,
                    cim_handler_props)
                if not cim_handler:
                    if "Filter" not in kwargs:
                        self._client.delete_instance(cim_filter.path)
                    lmi_raise_or_dump_exception(LMIIndicationError(errorstr))
                    return LMIReturnValue(rval=False, errorstr=errorstr)
            cim_subscription_props = {
                "Filter": cim_filter.path,
                "Handler": cim_handler.path
            }
            cim_subscription, _, errorstr = self._client.create_instance(
                kwargs.get(
                    "SubscriptionCreationClassName",
                    "CIM_IndicationSubscription"),
                indication_namespace,
                self.hostname,
                cim_subscription_props)
            if not cim_subscription:
                if "Filter" not in kwargs:
                    self._client.delete_instance(cim_filter.path)
                if "Handler" not in kwargs:
                    self._client.delete_instance(cim_handler.path)
                lmi_raise_or_dump_exception(LMIIndicationError(errorstr))
                return LMIReturnValue(rval=False, errorstr=errorstr)
            # XXX: Should we auto-delete all the indications?
            permanent = kwargs.get("Permanent", False)
            self._indications[kwargs["Name"]] = LMISubscription(
                self._client,
                (cim_filter, not cim_filter_provided),
                (cim_handler, not cim_handler_provided),
                cim_subscription,
                permanent)
        except KeyError, e:
            errorstr = "Not all necessary parameters provided, missing: %s" % e
            lmi_raise_or_dump_exception(LMIIndicationError(errorstr))
            return LMIReturnValue(rval=False, errorstr=errorstr)
        return LMIReturnValue(rval=True)

    def unsubscribe_indication(self, name):
        """
        Unsubscribes an indication.

        :param string name: indication name
        :returns: :py:class:`.LMIReturnValue` object with ``rval`` set to True,
            if unsubscribed; False otherwise
        """
        if name not in self._indications:
            errorstr = "No such indication"
            lmi_raise_or_dump_exception(LMIIndicationError(errorstr))
            return LMIReturnValue(rval=False, errorstr=errorstr)
        indication = self._indications.pop(name)
        indication.delete()
        return LMIReturnValue(rval=True)

    def unsubscribe_all_indications(self):
        """
        Unsubscribes all the indications. This call ignores *Permanent* flag,
        which may be provided in
        :py:meth:`.LMIConnection.subscribe_indication`, and deletes all the
        subscribed indications.
        """
        for subscription in self._indications.values():
            if not subscription.permanent:
                subscription.delete()
        self._indications = {}

    def print_subscribed_indications(self):
        """
        Prints out all the subscribed indications.
        """
        for i in self._indications.keys():
            sys.stdout.write("%s\n" % i)

    def subscribed_indications(self):
        """
        :returns: list of all the subscribed indications
        """
        return self._indications.keys()



# Copyright (C) 2013-2014 Michal Minar <miminar@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
"""
Meta classes simplifying declaration of user commands.

Each command is defined as a class with a set of properties. Some are
mandatory, the others have some default values. Each of them is transformed by
metaclasse to some function, class method or other property depending on
command type and semantic of property. Property itself is removed from
resulting class after being processed by meta class.
"""

import abc
import inspect
import re

from lmi.scripts.common import Configuration
from lmi.scripts.common import get_logger
from lmi.scripts.common import errors
from lmi.scripts.common.command import base
from lmi.scripts.common.command import util
from lmi.shell import LMIInstance
from lmi.shell.LMIReturnValue import LMIReturnValue

RE_CALLABLE = re.compile(
        r'^(?P<module>[a-z_]+(?:\.[a-z_]+)*):(?P<func>[a-z_]+)$',
        re.IGNORECASE)
RE_ARRAY_SUFFIX = re.compile(r'^(?:[a-z_]+[a-z0-9_]*)?$', re.IGNORECASE)
RE_OPTION = re.compile(r'^-+(?P<name>[^-+].*)$')
RE_MODULE_PATH = re.compile(r'([a-zA-z_]\w+\.)+[a-zA-z_]\w+')

FORMAT_OPTIONS = ('no_headings', 'human_friendly')

LOG = get_logger(__name__)

def _handle_usage(name, bases, dcl):
    """
    Take care of ``OWN_USAGE`` property. Supported values:

        `True`` :
            Means that documentation string of class is a usage string.
        ``False`` :
            No usage string for this command is defined.
        ``"usage string"`` :
            This property is a usage string.

    Defaults to ``False``.

    Usage string is an input parameter to ``docopt`` command-line options
    parser.

    :param string name: Name o command class.
    :param dictionary dcl: Class dictionary, which is modified by this
        function.
    """
    has_own_usage = False
    hlp = dcl.pop('OWN_USAGE', False)
    if hlp is True:
        if dcl['__doc__'] is None:
            raise errors.LmiCommandInvalidProperty(dcl['__module__'], name,
                    "OWN_USAGE set to True, but no __doc__ string present!")
        has_own_usage = True
    elif isinstance(hlp, basestring):
        if not '__doc__' in dcl:
            dcl['__doc__'] = hlp
        else:
            if not 'get_usage' in dcl:
                def _new_get_usage(_self, proper=False):
                    """ Get the usage string for ``doctopt`` parser. """
                    return hlp
                dcl['get_usage'] = _new_get_usage
        has_own_usage = True
    elif (   dcl.get('__node__', None) is None
         and any(getattr(b, 'has_own_usage', lambda: False)() for b in bases)):
        # inherit doc string of base class
        dcl['__doc__'] = (  b.__doc__ for b in bases
                         if getattr(b, 'has_own_usage', lambda: False)()).next()
        has_own_usage = True
    if has_own_usage:
        if not 'has_own_usage' in dcl:
            dcl['has_own_usage'] = classmethod(lambda _cls: True)

def _make_execute_method(bases, dcl, func):
    """
    Creates ``execute()`` method of a new end point command.

    :param tuple bases: Base classes of new command.
    :param dictionary dcl: Class dictionary being modified by this method.
    :param callable func: A callable wrapped by this new command. It's usually
        being referred to as *associated function*. If ``None``, no function
        will be created -- ``dcl`` won't be modified.
    """
    if func is not None and util.is_abstract_method(
            bases, 'execute', missing_is_abstract=True):
        del dcl['CALLABLE']
        def _execute(__self__, __connection__, *args, **kwargs):
            """ Invokes associated function with given arguments. """
            return func(__connection__, *args, **kwargs)
        _execute.dest = func
        dcl['execute'] = _execute

def _handle_namespace(dcl):
    """
    Overrides ``cim_namespace()`` class method if ``NAMESPACE`` property
    is given.

    :param dictionary dcl: Class dictionary being modified by this method.
    """
    if 'NAMESPACE' in dcl:
        namespace = dcl.pop('NAMESPACE')
        def _new_cim_namespace(_cls):
            """ Returns cim namespace used to modify connection object. """
            return namespace
        dcl['cim_namespace'] = classmethod(_new_cim_namespace)

def _handle_callable(name, bases, dcl):
    """
    Process the ``CALLABLE`` property of end-point command. Create the
    ``execute()`` method based on it.

    :param string name: Name of command class to create.
    :param tuple bases: Base classes of new command.
    :param dictionary dcl: Class dictionary being modified by this method.
    """
    try:
        func = dcl.get('CALLABLE')
        if isinstance(func, basestring):
            match = RE_CALLABLE.match(func)
            if not match:
                raise errors.LmiCommandInvalidCallable(
                        dcl['__module__'], name,
                        'Callable "%s" has invalid format (\':\' expected)'
                        % func)
            mod_name = match.group('module')
            try:
                func = getattr(__import__(mod_name, globals(), locals(),
                        [match.group('func')], 0),
                        match.group('func'))
            except (ImportError, AttributeError):
                raise errors.LmiImportCallableFailed(
                        dcl['__module__'], name, func)
    except KeyError:
        raise errors.LmiCommandMissingCallable(dcl['__module__'], name)
    if func is not None and not callable(func):
        raise errors.LmiCommandInvalidCallable(
            '"%s" is not a callable object or function.' % (
                func.__module__ + '.' + func.__name__))

    _make_execute_method(bases, dcl, func)

def _make_render_all_properties(bases):
    """
    Creates ``render()`` method, rendering all properties of instance.

    :param tuple bases: Base classes of new command class.
    :returns: Rendering method taking CIM instance as an
        argument.
    :rtype: function
    """
    if util.is_abstract_method(bases, 'render', missing_is_abstract=True):
        def _render(_self, inst):
            """
            Return tuple of ``(column_names, values)`` ready for output by
            formatter.
            """
            column_names, values = [], []
            for prop_name, value in sorted(inst.properties_dict().iteritems()):
                column_names.append(prop_name)
                if value is None:
                    value = ''
                values.append(value)
            return (column_names, values)

        return _render

def _make_render_with_properties(properties, target_formatter_lister=False):
    """
    Creates ``render()`` method, rendering given instance properties.

    :param properties: (``list``) List of properties to render.
    :param target_formatter_lister: (``bool``) Whether the output is targeted
        for Show command or Lister. The former expects a pair of column_names
        and values. The latter expects just values.
    :rtype: (``function``) Rendering method taking CIM instance as an
        argument.
    """
    def _process_property(prop, inst):
        """
        Takes a single property and instance. Returns computed value.

        :rtype: ``(str, any)`` A pair of property name and value.
        """
        if isinstance(prop, basestring):
            prop_name = prop
            if not prop in inst.properties():
                LOG().warn('Property "%s" not present in instance of "%s".',
                        prop, inst.path)
                value = "UNKNOWN"
            else:
                value = getattr(inst, prop)
        else:
            if not isinstance(prop, (tuple, list)):
                raise TypeError("prop must be a string or tuple, not %s" %
                        repr(prop))
            prop_name = prop[0]
            try:
                if callable(prop[1]):
                    value = prop[1](inst)
                else:
                    value = getattr(inst, prop[1])
            except Exception as exc:
                LOG().exception('Failed to render property "%s": %s',
                        prop[0], exc)
                value = "ERROR"
        if value is None:
            value = ''
        return prop_name, value

    if target_formatter_lister:
        def _render(self, inst):
            """
            Renders a limited set of properties and returns a row for instance
            table composed of property values.
            """
            if not isinstance(inst, LMIInstance):
                raise errors.LmiUnexpectedResult(
                        self.__class__, 'LMIInstance object', inst)
            return tuple(_process_property(p, inst)[1] for p in properties)

    else:
        def _render(self, inst):
            """
            Renders a limited set of properties and returns a pair of
            column names and values.
            """
            if not isinstance(inst, LMIInstance):
                raise errors.LmiUnexpectedResult(
                        self.__class__, 'LMIInstance object', inst)
            column_names, values = [], []
            for prop in properties:
                prop_name, value = _process_property(prop, inst)
                column_names.append(prop_name)
                values.append(value)
            return (column_names, values)

    return _render

def _check_render_properties(name, dcl, props):
    """
    Make sanity check for ``PROPERTIES`` class property. Exception will be
    raised when any flaw discovered.

    :param string name: Name of class to be created.
    :param dictionary dcl: Class dictionary.
    :param list props: List of properties or ``None``.
    """
    if props is not None:
        for prop in props:
            if not isinstance(prop, (basestring, tuple, list)):
                raise errors.LmiCommandInvalidProperty(
                        dcl['__module__'], name,
                        'PROPERTIES must be a list of strings or tuples')
            if isinstance(prop, (tuple, list)):
                if (  len(prop) != 2
                   or not isinstance(prop[0], basestring)
                   or (   not callable(prop[1])
                      and not isinstance(prop[1], basestring))):
                    raise errors.LmiCommandInvalidProperty(
                            dcl['__module__'], name,
                        'tuples in PROPERTIES must be: ("name",'
                        ' callable or property_name)')

def _handle_render_properties(name, bases, dcl, target_formatter_lister=False):
    """
    Process properties related to rendering function for commands operating
    on CIM instances. Result of this function a ``render()`` and
    ``get_columns()`` functions being added to class's dictionary with
    regard to handled properties.

    Currently handled properties are:

        ``DYNAMIC_PROPERTIES`` : ``bool``
            Whether the associated function itself provides list of
            properties. Optional property.
        ``PROPERTIES`` : ``bool``
            List of instance properties to print. Optional property.

    :param string name: Name of class to be created.
    :param tuple bases: Base classes of new command.
    :param dictionary dcl: Class dictionary being modified by this method.
    :param boolean target_formatter_lister: Whether the output is targeted
        for *Show* command or *Lister*. The former expects a pair of
        column_names and values. The latter expects just values.
    """
    dynamic_properties = dcl.pop('DYNAMIC_PROPERTIES', False)
    if dynamic_properties and 'PROPERTIES' in dcl:
        raise errors.LmiCommandError(
                dcl['__module__'], name,
                'DYNAMIC_PROPERTIES and PROPERTIES are mutually exclusive')

    properties = dcl.pop('PROPERTIES', None)
    _check_render_properties(name, dcl, properties)

    renderer = None
    get_columns = lambda cls: None
    if properties is None and not dynamic_properties:
        if (   target_formatter_lister
           and dcl.get('__metaclass__', None) is not InstanceListerMetaClass):
            raise errors.LmiCommandError(dcl['__module__'], name,
                    "either PROPERTIES must be declared or"
                    " DYNAMIC_PROPERTIES == True for InstanceLister"
                    " commands")
        renderer = _make_render_all_properties(bases)
    elif properties is None and dynamic_properties:
        def _render_dynamic(self, return_value):
            """ Renderer of dynamic properties. """
            properties, inst = return_value
            return _make_render_with_properties(properties,
                    target_formatter_lister)(self, inst)
        renderer = _render_dynamic
    elif properties is not None:
        renderer = _make_render_with_properties(properties,
                target_formatter_lister)
        get_columns = (lambda cls:
                    tuple((p[0] if isinstance(p, tuple) else p)
                for p in properties))
    if renderer is not None:
        dcl['render'] = classmethod(renderer)
    if target_formatter_lister:
        dcl['get_columns'] = get_columns

def _handle_opt_preprocess(name, dcl):
    """
    Process properties, that cause modification of parsed argument names before
    passing them to ``verify_options()`` or ``transform_options()``. If any of
    handled properties is supplied, it causes ``_preprocess_options()`` to be
    overriden, where all of desired name modifications will be made.
    Currently handled properties are:

        ``OPT_NO_UNDERSCORES`` : ``bool``
            When making a function's parameter name out of option, the leading
            dashes are replaced with underscore. If this property is True,
            dashes will be removed completely with no replacement.
        ``ARG_ARRAY_SUFFIX`` : ``bool``
            Add given suffix to all arguments resulting in list objects.

    :param string name: Command class name.
    :param dictionary dcl: Class dictionary being modified by this method.
    """
    if (   dcl.get('__metaclass__', None) is not EndPointCommandMetaClass
       and '_preprocess_options' in dcl):
        raise errors.LmiCommandError(dcl['__module__'], name,
                '_preprocess_options() method must not be overriden in the'
                'body of command class; use transform_options() instead')
    arr_suffix = dcl.pop('ARG_ARRAY_SUFFIX', '')
    if (  not isinstance(arr_suffix, str)
       or not RE_ARRAY_SUFFIX.match(arr_suffix)):
        raise errors.LmiCommandInvalidProperty(dcl['__module__'], name,
                'ARG_ARRAY_SUFFIX must be a string matching regular'
                ' expression "%s"' % RE_ARRAY_SUFFIX.pattern)
    opt_no_underscores = dcl.pop('OPT_NO_UNDERSCORES', False)
    if arr_suffix or opt_no_underscores:
        def _new_preprocess_options(_self, options):
            """ Modify (in-place) given options dictionary by renaming keys. """
            for do_it, cond, transform in (
                    ( arr_suffix
                    , lambda _, v: isinstance(v, list)
                    , lambda n   :
                              ('<' + util.RE_OPT_BRACKET_ARGUMENT.match(n)
                                    .group(1) + arr_suffix + '>')
                        if   util.RE_OPT_BRACKET_ARGUMENT.match(n)
                        else (n + arr_suffix))
                  , ( opt_no_underscores
                    , lambda n, _: RE_OPTION.match(n)
                    , lambda n   : RE_OPTION.match(n).group('name'))
                  ):
                if not do_it:
                    continue
                to_rename = (  name for name, value in options.items()
                            if cond(name, value))
                for name in to_rename:
                    new_name = transform(name)
                    LOG().debug('Renaming option "%s" to "%s".', name, new_name)
                    if new_name in options:
                        LOG().warn(
                            'Existing option named "%s" replaced with "%s".',
                                new_name, name)
                    options[new_name] = options.pop(name)

        dcl['_preprocess_options'] = _new_preprocess_options

def _handle_fallback_command(name, bases, dcl):
    """
    Process ``FALLBACK_COMMAND`` property of multiplexer command. It's turned
    into a :py:meth:`~.multiplexer.LmiCommandMultiplexer.fallback_command`
    class method. It needs to be called after the usage string is handled.

    .. seealso::
        :py:func:`_handle_usage`
    """
    fallback = dcl.pop('FALLBACK_COMMAND', None)
    if fallback is not None:
        if not issubclass(type(fallback), EndPointCommandMetaClass):
            raise errors.LmiCommandInvalidProperty(dcl['__module__'], name,
                    "FALLBACK_COMMAND must be a command class"
                    " (subclass of LmiEndPointCommand) not %s" % repr(fallback))
        if not fallback.has_own_usage():
            usage_string = dcl.get('__doc__', None)
            if not usage_string:
                for base_cls in bases:
                    if not issubclass(base_cls, base.LmiBaseCommand):
                        continue
                    cmd = base_cls
                    while not cmd.has_own_usage() and cmd.parent is not None:
                        cmd = cmd.parent
                    usage_string = cmd.__doc__
            if not usage_string:
                errors.LmiCommandError(dcl['__module__'], name,
                        "Missing usage string.")
            fallback.__doc__ = usage_string
            fallback.has_own_usage = lambda cls: True
        dcl['fallback_command'] = staticmethod(lambda: fallback)

def _handle_format_options(name, bases, dcl):
    """
    Process any ``FMT_*`` properties. This overrides ``format_options``
    property which returns dictionary of arguments passed to formatter factory.

    These properties are removed from class' dictionary.
    """
    format_options = {}
    for key, value in dcl.items():
        if key.startswith("FMT_"):
            opt_name = key[4:].lower()
            if opt_name not in FORMAT_OPTIONS:
                raise errors.LmiCommandInvalidProperty(
                        dcl['__module__'], name,
                        'Formatting option "%s" is not supported.' %
                        opt_name)
            if (   opt_name in ('no_headings', 'human_friendly')
               and not isinstance(value, bool)):
                raise errors.LmiCommandInvalidProperty(
                        dcl['__module__'], name,
                        '"%s" property must be a boolean')
            format_options[opt_name] = value

    if format_options:
        def _new_format_options(self):
            """ :returns: Dictionary of options for formatter object. """
            basecls = [b for b in bases if issubclass(b, base.LmiBaseCommand)][0]
            opts = basecls.format_options.fget(self)
            opts.update(format_options)
            return opts
        if 'format_options' in dcl:
            raise errors.LmiCommandError(dcl['__module__'], name,
                    'can not define both FMT_ options and "format_options" in'
                    ' the same class, choose just one of them')
        dcl['format_options'] = property(_new_format_options)

    for key in format_options:
        dcl.pop('FMT_' + key.upper())

def _handle_select(name, dcl):
    """
    Process properties of :py:class:`.select.LmiSelectCommand`.
    Currently handled properties are:

        ``SELECT`` : ``list``
            Is a list of pairs ``(condition, command)`` where ``condition`` is
            an expression in *LMIReSpL* language. And ``command`` is either a
            string with absolute path to command that shall be loaded or the
            command class itself.

            Small example: ::

                SELECT = [
                      ( 'OpenLMI-Hardware < 0.4.2'
                      , 'lmi.scripts.hardware.pre042.PreCmd'
                      )
                    , ('OpenLMI-Hardware >= 0.4.2 & class LMI_Chassis == 0.3.0'
                      , HwCmd
                      )
                ]

            It says: Let the ``PreHwCmd`` command do the job on brokers having
            ``openlmi-hardware`` package older than ``0.4.2``. Use the
            ``HwCmd`` anywhere else where also the ``LMI_Chassis`` CIM class in
            version ``0.3.0`` is available.

            First matching condition wins and assigned command will be passed
            all the arguments.

        ``DEFAULT`` : ``str`` or :py:class:`~.base.LmiBaseCommand`
            Defines fallback command used in case no condition can be
            satisfied.

    They will be turned into ``get_conditionals()`` method.
    """
    module_name = dcl.get('__module__', name)
    if not 'SELECT' in dcl:
        raise errors.LmiCommandError(module_name, name,
                "Missing SELECT property.")
    def inv_prop(msg, *args):
        return errors.LmiCommandInvalidProperty(module_name, name, msg % args)
    expressions = dcl.pop('SELECT')
    if not isinstance(expressions, (list, tuple)):
        raise inv_prop('SELECT must be list or tuple.')
    if len(expressions) < 1:
        raise inv_prop('SELECT must contain at least one condition!')
    for index, item in enumerate(expressions):
        if not isinstance(item, tuple):
            raise inv_prop('Items of SELECT must be tuples, not %s!' %
                    getattr(type(item), '__name__', 'UNKNOWN'))
        if len(item) != 2:
            raise inv_prop('Expected pair in SELECT on index %d!' % index)
        expr, cmd = item
        if not isinstance(expr, basestring):
            raise inv_prop('Expected expression string on index %d'
                    ' in SELECT!' % index)
        if isinstance(cmd, basestring) and not RE_MODULE_PATH.match(cmd):
            raise inv_prop('Second item of conditional pair on index %d'
                    ' in SELECT does not look as an importable path!' % cmd)
        if (   not isinstance(cmd, basestring)
           and not issubclass(cmd, (basestring, base.LmiBaseCommand))):
            raise inv_prop('Expected subclass of LmiBaseCommand (or its import'
                    ' path) as a second item of a pair on index %d in SELECT!'
                    % index)

    default = dcl.pop('DEFAULT', None)
    if isinstance(default, basestring) and not RE_MODULE_PATH.match(default):
        raise inv_prop('DEFAULT "%s" does not look as an importable path!'
            % default)
    if (   default is not None and not isinstance(default, basestring)
       and not issubclass(default, (basestring, base.LmiBaseCommand))):
        raise inv_prop('Expected subclass of LmiBaseCommand'
                ' (or its import path) as a value of DEFAULT!')
    def _new_get_conditionals(self):
        return expressions, default

    dcl['get_conditionals'] = _new_get_conditionals

class EndPointCommandMetaClass(abc.ABCMeta):
    """
    End point command does not have any subcommands. It's a leaf of
    command tree. It wraps some function in command library being
    referred to as an *associated function*. It handles following class
    properties:

        ``CALLABLE`` : ``str`` or callable
            An associated function. Mandatory property.
        ``OWN_USAGE`` : ``bool`` or ``str``
            Usage string. Optional property.
        ``ARG_ARRAY_SUFFIX`` : ``str``
            Suffix added to argument names containing array of values.
            Optional property.
        ``FMT_NO_HEADINGS`` : ``bool``
            Allows to force printing of table headers on and off for
            this command. Default is to print them.
        ``FMT_HUMAN_FRIENDLY`` : ``bool``
            Tells formatter to make the output more human friendly. The result
            is dependent on the type of formatter used.
    """

    def __new__(mcs, name, bases, dcl):
        _handle_usage(name, bases, dcl)
        _handle_callable(name, bases, dcl)
        _handle_opt_preprocess(name, dcl)
        _handle_format_options(name, bases, dcl)

        cls = super(EndPointCommandMetaClass, mcs).__new__(
                mcs, name, bases, dcl)

        # make additional check for arguments count
        dest = getattr(cls.execute, "dest", cls.execute)
        argspec = inspect.getargspec(dest)
        if (   not argspec.varargs
           and len(argspec.args) < cls.dest_pos_args_count()):
            raise errors.LmiCommandInvalidCallable(
                    dcl['__module__'], name,
                    'Callable must accept at least %d positional arguments' %
                    cls.dest_pos_args_count())

        return cls

class SessionCommandMetaClass(EndPointCommandMetaClass):
    """
    Meta class for commands operating upon a session object.
    All associated functions take as first argument an namespace abstraction
    of type ``lmi.shell``.

    Handles following class properties:

        ``NAMESPACE`` : ``str``
            CIM namespace abstraction that will be passed to associated
            function. Defaults to ``"root/cimv2"``. If ``False``, raw
            :py:class:`lmi.shell.LMIConnection` object will be passed to
            associated function.
    """
    def __new__(mcs, name, bases, dcl):
        _handle_usage(name, bases, dcl)
        _handle_namespace(dcl)
        _handle_callable(name, bases, dcl)

        return EndPointCommandMetaClass.__new__(mcs, name, bases, dcl)

class ListerMetaClass(SessionCommandMetaClass):
    """
    Meta class for end-point lister commands. Handles following class
    properties:

        ``COLUMNS`` : ``tuple``
            List of column names. Optional property. There are special values
            such as:

                ``None`` or omitted
                    Associated function provides column names in a first row of
                    returned list or generator.

                empty list, empty tuple or ``False``
                    They mean that no headers shall be printed. It is simalar
                    to using ``FMT_NO_HEADINGS = True``. But in this case all
                    the rows returned from associated functions are treated as
                    data.
    """

    def __new__(mcs, name, bases, dcl):
        cols = dcl.pop('COLUMNS', None)
        if cols is not None:
            if not isinstance(cols, (list, tuple)):
                raise errors.LmiCommandInvalidProperty(dcl['__module__'], name,
                        'COLUMNS class property must be either list or tuple')
            if len(cols) < 1 or cols is False:
                dcl['FMT_NO_HEADINGS'] = True
                cols = tuple()
            elif not all(isinstance(c, basestring) for c in cols):
                raise errors.LmiCommandInvalidProperty(dcl['__module__'], name,
                        'COLUMNS must contain just column names as strings')
            def _new_get_columns(_cls):
                """ Return column names. """
                return cols
            dcl['get_columns'] = classmethod(_new_get_columns)

        return super(ListerMetaClass, mcs).__new__(mcs, name, bases, dcl)

class ShowInstanceMetaClass(SessionCommandMetaClass):
    """
    Meta class for end-point show instance commands. Additional handled
    properties:

        ``DYNAMIC_PROPERTIES`` : ``bool``
            Whether the associated function itself provides list of
            properties. Optional property.
        ``PROPERTIES`` : ``tuple``
            List of instance properties to print. Optional property.

    These are translated in a :py:meth:`~.show.LmiShowInstance.render`, which
    should be marked as abstract in base lister class.
    """

    def __new__(mcs, name, bases, dcl):
        _handle_render_properties(name, bases, dcl)

        return super(ShowInstanceMetaClass, mcs).__new__(
                mcs, name, bases, dcl)

class InstanceListerMetaClass(SessionCommandMetaClass):
    """
    Meta class for instance lister command handling the same properties
    as :py:class:`ShowInstanceMetaClass`.
    """

    def __new__(mcs, name, bases, dcl):
        _handle_render_properties(name, bases, dcl, True)

        return super(InstanceListerMetaClass, mcs).__new__(
                mcs, name, bases, dcl)

class CheckResultMetaClass(SessionCommandMetaClass):
    """
    Meta class for end-point command "check result". Additional handled
    properties:

        ``EXPECT`` :
            Value to compare against the return value. Mandatory property.

    ``EXPECT`` property is transformed into a
    :py:meth:`.checkresult.LmiCheckResult.check_result` method taking two
    arguments ``(options, result)`` and returning a boolean.
    """

    def __new__(mcs, name, bases, dcl):
        try:
            expect = dcl['EXPECT']
            if callable(expect):
                def _new_expect(_self, options, result):
                    """
                    Comparison function testing return value with *expect*
                    function.
                    """
                    if isinstance(result, LMIReturnValue):
                        result = result.rval
                    passed = expect(options, result)
                    if not passed:
                        LOG().info('Got unexpected result "%s".')
                    return passed
            else:
                def _new_expect(_self, _options, result):
                    """ Comparison function testing by equivalence. """
                    if isinstance(result, LMIReturnValue):
                        result = result.rval
                    passed = expect == result
                    if not passed:
                        LOG().info('Expected "%s", got "%s".', expect, result)
                        return (False, '%s != %s' % (expect, result))
                    return passed
                _new_expect.expected = expect
            del dcl['EXPECT']
            dcl['check_result'] = _new_expect
        except KeyError:
            # EXPECT might be defined in some subclass
            pass

        return super(CheckResultMetaClass, mcs).__new__(mcs, name, bases, dcl)

class MultiplexerMetaClass(abc.ABCMeta):
    """
    Meta class for node command (not an end-point command). It handles
    following class properties:

        ``COMMANDS`` : ``dict``
            Command names with assigned command classes. Each of them is a
            direct subcommands of command with this property. Mandatory
            property.

        ``FALLBACK_COMMAND`` : :py:class:`~.endpoint.LmiEndPointCommand`
            Command factory to use in case that no command is passed on command
            line.

    Formatting options (starting with ``FMT_`` are also accepted, and may used
    to set defaults for all subcommands.
    """

    def __new__(mcs, name, bases, dcl):
        if dcl.get('__metaclass__', None) is not MultiplexerMetaClass:
            module_name = dcl.get('__module__', name)
            # check COMMANDS property and make it a classmethod
            if not 'COMMANDS' in dcl:
                raise errors.LmiCommandError(module_name, name,
                        'Missing COMMANDS property.')
            cmds = dcl.pop('COMMANDS')
            if not isinstance(cmds, dict):
                raise errors.LmiCommandInvalidProperty(module_name, name,
                        'COMMANDS must be a dictionary')
            if not all(isinstance(c, basestring) for c in cmds.keys()):
                raise errors.LmiCommandInvalidProperty(module_name, name,
                        'Keys of COMMANDS dictionary must contain command'
                        ' names as strings.')
            for cmd_name, cmd in cmds.items():
                if not util.RE_COMMAND_NAME.match(cmd_name):
                    raise errors.LmiCommandInvalidName(
                            module_name, name, cmd_name)
                if not issubclass(cmd, base.LmiBaseCommand):
                    raise errors.LmiCommandError(module_name, name,
                            'COMMANDS dictionary must be composed of'
                            ' LmiBaseCommand subclasses, failed class: "%s"'
                            % cmd.__name__)
                if cmd.is_multiplexer() and not cmd.has_own_usage():
                    LOG().warn('Command "%s.%s" is missing usage string.'
                            ' It will be inherited from parent command.',
                            cmd.__module__, cmd.__name__)
                    cmd.__doc__ = dcl['__doc__']
            def _new_child_commands(_cls):
                """ Returns list of subcommands. """
                return cmds
            dcl['child_commands'] = classmethod(_new_child_commands)

            _handle_usage(name, bases, dcl)
            _handle_fallback_command(name, bases, dcl)
            _handle_format_options(name, bases, dcl)

        return super(MultiplexerMetaClass, mcs).__new__(mcs, name, bases, dcl)

class SelectMetaClass(abc.ABCMeta):
    """
    Meta class for select commands with guarded commands. Additional handled
    properties:

        ``SELECT`` : ``list``
            List of commands guarded with expressions representing requirements
            on server's side that need to be met.
        ``DEFAULT`` : ``str`` or :py:class:`~.base.LmiBaseCommand`
            Defines fallback command used in case no condition can is
            satisfied.
    """

    def __new__(mcs, name, bases, dcl):
        if dcl.get('__metaclass__', None) is not SelectMetaClass:
            module_name = dcl.get('__module__', name)
            if not '__doc__' in dcl:
                LOG().warn('Command selector "%s.%s" is missing short'
                    ' description string (__doc__).',
                        module_name, name)
                default = dcl.get('DEFAULT', None)
                if (    default is not None
                        and issubclass(default, base.LmiBaseCommand)
                        and getattr(dcl['DEFAULT'], '__doc__', None)):
                    LOG().warn('Using __doc__ string from default command for'
                        ' selector "%s.%s".', module_name, name)
                    dcl['__doc__'] = dcl['DEFAULT'].__doc__
            _handle_select(name, dcl)
        return super(SelectMetaClass, mcs).__new__(mcs, name, bases, dcl)




# Copyright (C) 2012-2014 Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import collections

from lmi.shell.compat import *

_RValue = collections.namedtuple('RValue', ["rval", "rparams", "errorstr"])


class LMIReturnValue(_RValue):
    """
    Class representing a return value, which holds 3 main types of attributes:

    :param rval: return value
    :param dictionary rparams: returned parameters of e.g. method call
    :param string errorstr: error string
    """
    def __repr__(self):
        """
        :returns: pretty string
        """
        return "LMIReturnValue(rval=%r, rparams=%r, errorstr=%r)" % self

    def __new__(self, rval, rparams=None, errorstr=None):
        rparams = rparams if rparams else {}
        if not isinstance(rparams, wbem.NocaseDict):
            rparams = wbem.NocaseDict(rparams)
        errorstr = errorstr if errorstr else ""
        return _RValue.__new__(
            self, rval=rval, rparams=rparams, errorstr=errorstr)



# Copyright (C) 2012-2014 Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import os
import subprocess
import sys
from setuptools import setup, find_packages

long_description = ""
try:
    try:
        script_dir = os.path.dirname(sys.argv[0])
        cmd = ['/usr/bin/make', 'readme']
        readme_file = 'README.txt'
        if script_dir not in (',', ''):
            cmd[1:1] = ['-C', script_dir]
            readme_file = os.path.join(script_dir, readme_file)
        with open('/dev/null', 'w') as null:
            ret = subprocess.call(cmd, stdout=null, stderr=null)
        if not ret:
            long_description = open(readme_file, 'rt').read()
    except Exception as err:
        sys.stderr.write('ERROR while reading README.txt: %s\n' % str(err))
    if not long_description:
        long_description = open('README.md', 'rt').read()
except IOError:
    pass

setup(
    name="openlmi-tools",
    description="OpenLMI (non)interactive shell and meta-command",
    long_description=long_description,
    version='0.10.1',
    license="GPLv2+",
    url="http://fedorahosted.org/openlmi/",
    author="Peter Hatina, Michal Minar",
    author_email="phatina@redhat.com, miminar@redhat.com",
    classifiers=[
        "License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)",
        "Operating System :: POSIX :: Linux",
        "Topic :: System :: Systems Administration",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.2",
        "Intended Audience :: Developers",
        "Environment :: Console",
    ],

    namespace_packages=["lmi", "lmi.scripts"],
    packages=(
            [ "lmi"
            , 'lmi.shell'
            , 'lmi.shell.compat'
            , 'lmi.scripts'
            , 'lmi.scripts._metacommand'
            , 'lmi.scripts.common'
            , 'lmi.scripts.common.command'
            , 'lmi.scripts.common.formatter'
            , 'lmi.scripts.common.versioncheck']),
    install_requires=["docopt >= 0.6", "openlmi", "pyparsing"],
    scripts=["lmishell"],

    entry_points={
        "console_scripts": [
            "lmi = lmi.scripts._metacommand:main"
            ],
        "lmi.scripts.cmd": [],
        },
    zip_safe=False,
)



# Copyright (C) 2013-2014 Michal Minar <miminar@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
"""
Utility functions used in command sub-package.
"""

import inspect
import os
import re

#: Regular expression matching bracket argument such as ``<arg_name>``.
RE_OPT_BRACKET_ARGUMENT = re.compile('^<(?P<name>[^>]+)>$')
#: Regular expression matching argument written in upper case such as
#:``ARG_NAME``.
RE_OPT_UPPER_ARGUMENT = re.compile('^(?P<name>[A-Z0-9]+(?:[_-][A-Z0-9]+)*)$')
#: Regular expression matching showt options. They are one character
#: long, prefixed with single dash.
RE_OPT_SHORT_OPTION = re.compile('^-(?P<name>[a-z])$', re.IGNORECASE)
#: Regular expression matching long options (prefixed with double dash).
RE_OPT_LONG_OPTION = re.compile('^--(?P<name>[a-z0-9_-]+)$', re.IGNORECASE)
#: Command name can also be a single or double dash.
RE_COMMAND_NAME = re.compile(r'^([a-z]+(-[a-z0-9]+)*|--?)$')

def is_abstract_method(clss, method, missing_is_abstract=False):
    """
    Check, whether the given method is abstract in given class or list of
    classes. May be used to check, whether we should override particular
    abstract method in a meta-class in case that no non-abstract
    implementation is defined.

    :param clss: Class or list of classes that is
        searched for non-abstract implementation of particular method.
        If the first class having particular method in this list contain
        non-abstract implementation, ``False`` is returned.
    :type clss: type or tuple
    :param string method: Name of method to look for.
    :param boolean missing_is_abstract: This is a value returned, when
        not such method is defined in a set of given classes.
    :returns: Are all occurences of given method abstract?
    :rtype: boolean
    """
    if (   not isinstance(clss, (list, tuple, set))
       and not isinstance(clss, type)):
        raise TypeError("clss must be either a class or a tuple of classes")
    if not isinstance(method, basestring):
        raise TypeError("method must be a string")
    if isinstance(clss, type):
        clss = [clss]
    for cls in clss:
        if hasattr(cls, method):
            if getattr(getattr(cls, method), "__isabstractmethod__", False):
                return True
            else:
                return False
    return missing_is_abstract

def get_module_name(frame_level=2):
    """
    Get a module name of caller from particular outer frame.

    :param integer frame_level: Number of nested frames to skip when searching
        for called function scope by inspecting stack upwards. When the result
        of this function is applied directly on the definition of function,
        it's value should be 1. When used from inside of some other factory, it
        must be increased by 1.

        Level 0 returns name of this module. Level 1 returns module name of
        caller. Level 2 returns module name of caller's caller.
    :returns: Module name.
    :rtype: string
    """
    frame = inspect.currentframe()
    while frame_level > 0 and frame.f_back:
        frame = frame.f_back
        frame_level -= 1
    module = getattr(frame, 'f_globals', {}).get('__name__', None)
    if module is None:
        if hasattr(frame, 'f_code'):
            module = os.path.basename(frame.f_code.co_filename.splitext())[0]
        else:
            module = '_unknown_'
    return module



# Copyright (C) 2013-2014 Red Hat, Inc. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and documentation are
# those of the authors and should not be interpreted as representing official
# policies, either expressed or implied, of the FreeBSD Project.
#
# Authors: Michal Minar <miminar@redhat.com>
#
"""
Package with utilities for checking availability of profiles or CIM classes.
Version requirements can also be specified.
"""

import functools
from pyparsing import ParseException

from lmi.scripts.common import Configuration
from lmi.scripts.common import errors
from lmi.scripts.common.versioncheck import parser

def cmp_profiles(fst, snd):
    """
    Compare two profiles by their version.

    :returns:
        * -1 if the *fst* profile has lower version than *snd*
        * 0  if their versions are equal
        * 1  otherwise
    :rtype: int
    """
    fstver = fst.RegisteredVersion
    sndver = snd.RegisteredVersion
    if fstver == sndver:
        return 0
    return -1 if parser.cmp_version(fstver, sndver) else 1

def get_profile_version(conn, name, cache=None):
    """
    Get version of registered profile on particular broker. Queries
    ``CIM_RegisteredProfile`` and ``CIM_RegisteredSubProfile``. The latter
    comes in question only when ``CIM_RegisteredProfile`` does not yield any
    matching result.

    :param conn: Connection object.
    :param string name: Name of the profile which must match value of *RegisteredName*
        property.
    :param dictionary cache: Optional cache where the result will be stored for
        later use. This greatly speeds up evaluation of several expressions refering
        to same profiles or classes.
    :returns: Version of matching profile found. If there were more of them,
        the highest version will be returned. ``None`` will be returned when no matching
        profile or subprofile is found.
    :rtype: string
    """
    if cache and name in cache:
        return cache[(conn.uri, name)]
    insts = conn.root.interop.wql('SELECT * FROM CIM_RegisteredProfile'
            ' WHERE RegisteredName=\"%s\"' % name)
    regular = set(i for i in insts if i.classname.endswith('RegisteredProfile'))
    if regular:     # select instances of PG_RegisteredProfile if available
        insts = regular
    else:           # otherwise fallback to PG_RegisteredSubProfile instances
        insts = set(i for i in insts if i not in regular)
    if not insts:
        ret = None
    else:
        ret = sorted(insts, cmp=cmp_profiles)[-1].RegisteredVersion
    if cache is not None:
        cache[(conn.uri, name)] = ret
    return ret

def get_class_version(conn, name, namespace=None, cache=None):
    """
    Query broker for version of particular CIM class. Version is stored in
    ``Version`` qualifier of particular CIM class.

    :param conn: Connection object.
    :param string name: Name of class to query.
    :param string namespace: Optional CIM namespace. Defaults to configured namespace.
    :param dictionary cache: Optional cache used to speed up expression prrocessing.
    :returns: Version of CIM matching class. Empty string if class is registered but
        is missing ``Version`` qualifier and ``None`` if it is not registered.
    :rtype: string
    """
    if namespace is None:
        namespace = Configuration.get_instance().namespace
    if cache and (namespace, name) in cache:
        return cache[(conn.uri, namespace, name)]
    ns = conn.get_namespace(namespace)
    cls = getattr(ns, name, None)
    if not cls:
        ret = None
    else:
        quals = cls.wrapped_object.qualifiers
        if 'Version' not in quals:
            ret = ''
        else:
            ret = quals['Version'].value
    if cache is not None:
        cache[(conn.uri, namespace, name)] = ret
    return ret

def eval_respl(expr, conn, namespace=None, cache=None):
    """
    Evaluate LMIReSpL expression on particular broker.

    :param string expr: Expression to evaluate.
    :param conn: Connection object.
    :param string namespace: Optional CIM namespace where CIM classes will be
        searched.
    :param dictionary cache: Optional cache speeding up evaluation.
    :returns: ``True`` if requirements in expression are satisfied.
    :rtype: boolean
    """
    if namespace is None:
        namespace = Configuration.get_instance().namespace
    stack = []
    pvget = functools.partial(get_profile_version, conn, cache=cache)
    cvget = functools.partial(get_class_version, conn,
            namespace=namespace, cache=cache)
    pr = parser.bnf_parser(stack, pvget, cvget)
    pr.parseString(expr, parseAll=True)
    # Now evaluate starting non-terminal created on stack.
    return stack[0]()




# Copyright (C) 2013-2014 Michal Minar <miminar@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
"""
Module with predefined exceptions for use in scripts.
"""

try:
    import lmiwbem as wbem
except ImportError:
    import pywbem as wbem

CIMError = wbem.CIMError

class LmiError(Exception):
    """
    The base Lmi scripts error.
    All the other exceptions inherit from it.
    """
    pass

class LmiFailed(LmiError):
    """
    Raised, when operation on remote host failes.
    It's supposed to be used especially in command libraries.
    """
    pass

class LmiUnsatisfiedDependencies(LmiFailed):
    """
    Raised when no guarded command in
    :py:class:`~.command.select.LmiSelectCommand` can be loaded due to
    unsatisfied requirements.
    """
    def __init__(self, uris):
        LmiFailed.__init__(self, "Profile and class dependencies were not"
                " satisfied for this session (%s)." % ', '.join(uris))

class LmiUnexpectedResult(LmiError):
    """
    Raised, when command's associated function returns something unexpected.
    """
    def __init__(self, command_class, expected, result):
        LmiError.__init__(self,
                'Got unexpected result from associated function of'
                ' "%s.%s", expected "%s", got: "%s".' %
                (command_class.__module__, command_class.__name__,
                    expected, repr(result)))

class LmiInvalidOptions(LmiError):
    """
    Raised in :py:meth:`~.command.endpoint.LmiEndPointCommand.verify_options`
    method if the options given are not valid.
    """
    pass

class LmiCommandNotFound(LmiError):
    """ Raised, when user requests not registered command. """
    def __init__(self, cmd_name):
        LmiError.__init__(self, 'No such command "%s".' % cmd_name)

class LmiNoConnections(LmiError):
    """ Raised, when no connection to remote hosts could be made. """
    pass

class LmiCommandImportError(LmiError):
    """ Exception raised when command can not be imported. """
    def __init__(self, cmd_name, cmd_path, reason):
        LmiError.__init__(self, 'Failed to import command "%s" (%s): %s' % (
                cmd_name, cmd_path, reason))

class LmiCommandError(LmiError):
    """ Generic exception related to command declaration. """
    def __init__(self, module_name, class_name, msg):
        LmiError.__init__(self, 'Wrong declaration of command "%s": %s'
                % (    ".".join([module_name, class_name])
                    if module_name else class_name
                  , msg))

class LmiCommandInvalidName(LmiCommandError):
    """ Raised, when command gets invalid name. """
    def __init__(self, module_name, class_name, cmd_name):
        LmiCommandError.__init__(self, module_name, class_name,
                'Invalid command name "%s".' % cmd_name)

class LmiCommandMissingCallable(LmiCommandError):
    """ Raised, when command declaration is missing callable object. """
    def __init__(self, module_name, class_name):
        LmiCommandError.__init__(self, module_name, class_name,
                'Missing CALLABLE property.')

class LmiCommandInvalidProperty(LmiCommandError):
    """ Raised, when any command property contains unexpected value. """
    pass

class LmiImportCallableFailed(LmiCommandInvalidProperty):
    """ Raised, when callable object of command could not be imported. """
    def __init__(self, module_name, class_name, callable_prop):
        LmiCommandInvalidProperty.__init__(self, module_name, class_name,
            'Failed to import callable "%s".' % callable_prop)

class LmiCommandInvalidCallable(LmiCommandInvalidProperty):
    """ Raised, when given callback is not callable. """
    def __init__(self, module_name, class_name, msg):
        LmiCommandInvalidProperty.__init__(self, module_name, class_name, msg)

class LmiBadSelectExpression(LmiCommandError):
    """
    Raised, when expression of :py:class:`~.command.select.LmiSelectCommand`
    could not be evaluated.
    """
    def __init__(self, module_name, class_name, expr):
        LmiCommandError.__init__(self, module_name, class_name,
                "Bad select expression: %s" % expr)

class LmiTerminate(Exception):
    """
    Raised to cleanly terminate interavtive shell.
    """
    def __init__(self, exit_code=0):
        Exception.__init__(self, exit_code)



# Copyright (C) 2013-2014 Michal Minar <miminar@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
"""
Module for Configuration class.

Configuration
---------------------

.. autoclass:: Configuration
    :members:

"""

import os
from lmi.base.BaseConfiguration import BaseConfiguration

LISTER_FORMATS = ['csv', 'table']

#: Default format string to use in stderr handlers.
DEFAULT_FORMAT_STRING = "%(cseq)s%(levelname_)-8s:%(creset)s %(message)s"

class Configuration(BaseConfiguration):
    """
    Configuration class specific to software providers.
    *OpenLMI* configuration file should reside in: ::

        /etc/openlmi/scripts/lmi.conf

    :param string user_config_file_path: Path to the user configuration
        options.
    """

    USER_CONFIG_FILE_PATH = "~/.lmirc"
    HISTORY_FILE = "~/.lmi_history"

    OUTPUT_SILENT  = -1
    OUTPUT_WARNING = 0
    OUTPUT_INFO    = 1
    OUTPUT_DEBUG   = 2

    # indexes to LISTER_FORMATS
    LISTER_FORMAT_CSV = 0
    LISTER_FORMAT_TABLE = 1

    def __init__(self, user_config_file_path=USER_CONFIG_FILE_PATH, **kwargs):
        self._user_config_file_path = os.path.expanduser(user_config_file_path)
        BaseConfiguration.__init__(self, **kwargs)
        self._verbosity = None
        self._trace = None
        self._verify_server_cert = None
        self._cim_namespace = None
        self._human_friendly = None
        self._lister_format = None
        self._no_headings = None
        self._log_file = None
        self._history_max_length = None

    @classmethod
    def provider_prefix(cls):
        return "scripts"

    @classmethod
    def default_options(cls):
        """
        :returns: Dictionary of default values.
        :rtype: dictionary
        """
        defaults = BaseConfiguration.default_options().copy()
        # [Main] options
        defaults["CommandNamespace"] = 'lmi.scripts.cmd'
        defaults["Trace"] = "False"
        defaults["Verbosity"] = "0"
        defaults["HistoryMaxLength"] = "4000"
        # [Log] options
        defaults['ConsoleFormat'] = DEFAULT_FORMAT_STRING
        defaults['ConsoleInfoFormat'] = '%(message)s'
        defaults['FileFormat'] = \
                "%(asctime)s:%(levelname)-8s:%(name)s:%(lineno)d - %(message)s"
        defaults['LogToConsole'] = 'True'
        defaults['OutputFile'] = ''
        # [SSL] options
        defaults['VerifyServerCertificate'] = 'True'
        # [Format] options
        defaults['HumanFriendly'] = 'False' # be ugly by default
        defaults['ListerFormat'] = 'table'
        defaults['NoHeadings'] = 'False'
        return defaults

    @classmethod
    def mandatory_sections(cls):
        sects = set(BaseConfiguration.mandatory_sections())
        sects.add('Main')
        sects.add('SSL')
        sects.add('Format')
        return list(sects)

    def load(self):
        """ Read additional user configuration file if it exists. """
        BaseConfiguration.load(self)
        self.config.read(self._user_config_file_path)

    # *************************************************************************
    # [CIM] options
    # *************************************************************************
    @property
    def namespace(self):
        if self._cim_namespace is None:
            return BaseConfiguration.namespace.fget(self)
        return self._cim_namespace
    @namespace.setter
    def namespace(self, namespace):
        if not isinstance(namespace, basestring) and namespace is not None:
            raise TypeError("namespace must be a string")
        self._cim_namespace = namespace

    # *************************************************************************
    # [Main] options
    # *************************************************************************
    @property
    def history_file(self):
        """ Path to a file with history of interactive mode. """
        return os.path.expanduser(self.HISTORY_FILE)

    @property
    def history_max_length(self):
        """ Maximum number of lines kept in history file. """
        return self.get_safe('Main', 'HistoryMaxLength', int)

    @property
    def silent(self):
        """ Whether to suppress all output messages except for errors. """
        return self.verbosity <= self.OUTPUT_SILENT

    @property
    def trace(self):
        """ Whether the tracebacks shall be printed upon errors. """
        if self._trace is not None:
            return self._trace
        return self.get_safe('Main', 'Trace', bool)

    @trace.setter
    def trace(self, trace):
        """ Allow to override configuration option Trace. """
        if trace is not None:
            trace = bool(trace)
        self._trace = trace

    @property
    def verbose(self):
        """ Whether to output more details. """
        return self.verbosity >= self.OUTPUT_INFO

    @property
    def verbosity(self):
        """ Return integer indicating verbosity level of output to console. """
        if self._verbosity is None:
            return self.get_safe('Main', 'Verbosity', int)
        return self._verbosity

    @verbosity.setter
    def verbosity(self, level):
        """ Allow to set verbosity without modifying configuration values. """
        if not isinstance(level, (long, int)) and level is not None:
            raise TypeError("level must be integer")
        if level is not None:
            if level < self.OUTPUT_SILENT:
                level = self.OUTPUT_SILENT
            elif level > self.OUTPUT_DEBUG:
                level = self.OUTPUT_DEBUG
        self._verbosity = level

    # *************************************************************************
    # [Log] options
    # *************************************************************************
    @property
    def log_file(self):
        """ Path to a file, where logging messages shall be written. """
        if self._log_file is None:
            return self.get_safe('Log', 'OutputFile')
        return self._log_file
    @log_file.setter
    def log_file(self, log_file):
        """ Override logging file path. """
        if log_file is not None and not isinstance(log_file, basestring):
            raise TypeError("log_file must be a string")
        self._log_file = log_file

    # *************************************************************************
    # [SSL] options
    # *************************************************************************
    @property
    def verify_server_cert(self):
        """
        Return boolean saying, whether the server-side certificate should be
        checked.
        """
        if self._verify_server_cert is None:
            return self.get_safe('SSL', 'VerifyServerCertificate', bool)
        return self._verify_server_cert
    @verify_server_cert.setter
    def verify_server_cert(self, value):
        """ Allows to override configuration option value. """
        if value is not None:
            value = bool(value)
        self._verify_server_cert = value

    # *************************************************************************
    # [Format] options
    # *************************************************************************
    @property
    def human_friendly(self):
        """ Whether to print human-friendly values. """
        if self._human_friendly is None:
            return self.get_safe('Format', 'HumanFriendly', bool)
        return self._human_friendly
    @human_friendly.setter
    def human_friendly(self, value):
        """ Allows to override configuration option value. """
        if value is not None:
            value = bool(value)
        self._human_friendly = value

    @property
    def lister_format(self):
        """
        Output format used for lister commands. Returns one of
            * LISTER_FORMAT_CSV
            * LISTER_FORMAT_TABLE

        :rtype: integer
        """
        if self._lister_format is None:
            value = self.get_safe('Format', 'ListerFormat')
            try:
                return LISTER_FORMATS.index(value.lower())
            except ValueError:
                value = self.default_options()['ListerFormat']
                return LISTER_FORMATS.index(value.lower())
        return self._lister_format
    @lister_format.setter
    def lister_format(self, value):
        """
        Allows to override configuration option.

        :param value: One of items from ``LISTER_FORMATS`` array or an index
            to it.
        :type value: integer or string
        """
        if (   value is not None
           and (  not isinstance(value, int)
               or (value < 0 or value >= len(LISTER_FORMATS)))
           and (  not isinstance(value, basestring)
               or value.lower() not in LISTER_FORMATS)) :
            raise TypeError("value must be an integer or one of: %s" %
                    LISTER_FORMATS)
        if isinstance(value, basestring):
            value = LISTER_FORMATS.index(value.lower())
        self._lister_format = value

    @property
    def no_headings(self):
        """ Whether to print headings of tables. """
        if self._no_headings is None:
            return self.get_safe('Format', 'NoHeadings', bool)
        return self._no_headings
    @no_headings.setter
    def no_headings(self, value):
        """ Allows to override configuration option. """
        if value is not None:
            value = bool(value)
        self._no_headings = value

# There were some path changes in BaseConfiguration after 0.2.0 release.
# Let's fallback to older variable name when the new one is not present.
if hasattr(BaseConfiguration, 'CONFIG_DIRECTORY_TEMPLATE_PROVIDER'):
    setattr( Configuration
           , 'CONFIG_FILE_PATH_TEMPLATE_PROVIDER'
           , getattr(BaseConfiguration, 'CONFIG_DIRECTORY_TEMPLATE_PROVIDER')
                + 'lmi.conf')
else:   # fallback
    setattr( Configuration
           , 'CONFIG_FILE_PATH_TEMPLATE'
           , getattr(BaseConfiguration, 'CONFIG_DIRECTORY_TEMPLATE')
                + 'lmi.conf')



# Copyright (C) 2013-2014 Red Hat, Inc. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and documentation are
# those of the authors and should not be interpreted as representing official
# policies, either expressed or implied, of the FreeBSD Project.
#
# Authors: Michal Minar <miminar@redhat.com>
#
"""
Defines command class used to choose other commands depending on
profile and class requirements.
"""

from docopt import docopt
from pyparsing import ParseException

from lmi.scripts.common import get_logger
from lmi.scripts.common import errors
from lmi.scripts.common.command import base
from lmi.scripts.common.command import meta
from lmi.scripts.common.session import SessionProxy
from lmi.scripts.common.versioncheck import eval_respl

class LmiSelectCommand(base.LmiBaseCommand):
    """
    Base class for command selectors. It does not process command line
    arguments. Thery are passed unchanged to selected command whose
    requirements are met. Its doc string is not interpreted in any way.

    If there are more hosts, conditions are evaluated per each. They are then
    split into groups, each fulfilling particular condition. Associated
    commands are then invoked on these groups separately.

    Example usage: ::

        class MySelect(LmiSelectCommand):
            SELECT = [
                ( 'OpenLMI-Hardware >= 0.4.2'
                , 'lmi.scripts.hardware.current.Cmd'),
                ('OpenLMI-Hardware', 'lmi.scripts.hardware.pre042.Cmd')
            ]
            DEFAULT = MissingHwProviderCmd

    Using metaclass: :py:class:`.meta.SelectMetaClass`.
    """
    __metaclass__ = meta.SelectMetaClass

    @classmethod
    def is_end_point(cls):
        return False

    @classmethod
    def is_multiplexer(cls):
        return False

    @classmethod
    def get_conditionals(cls):
        """
        Get the expressions with associated commands. This shall be overriden
        by a subclass.

        :returns: Pair of ``(expressions, default)``.
            Where ``expressions`` is a list of pairs ``(condition, command)``.
            And ``default`` is the same as ``command`` used in case no
            ``condition`` is satisfied.
        :rtype: list
        """
        raise NotImplementedError(
                "get_conditionals needs to be defined in subclass")

    def eval_expr(self, expr, hosts, cache=None):
        """
        Evaluate expression on group of hosts.

        :param string expr: Expression to evaluate.
        :param list hosts: Group of hosts that shall be checked.
        :param dictionary cache: Optional cache object speeding up evaluation
            by reducing number of queries to broker.
        :returns: Subset of hosts satisfying *expr*.
        :rtype: list
        """
        if cache is None:
            cache = dict()
        session = self.session
        satisfied = []
        try:
            for host in hosts:  # TODO: could be done concurrently
                conn = session[host]
                if not conn:
                    continue
                if eval_respl(expr, conn, cache=cache):
                    satisfied.append(host)
        except ParseException:
            raise errors.LmiBadSelectExpression(self.__class__.__module__,
                    self.__class__.__name__, "Bad select expression: %s" % expr)
        return satisfied

    def select_cmds(self, cache=None):
        """
        Generator of command factories with associated groups of hosts. It
        evaluates given expressions on session. In this process all expressions
        from :py:meth:`get_conditionals` are checked in a row. Host satisfying
        some expression is added to group associated with it and is excluded
        from processing following expressions.

        :param dictionary cache: Optional cache object speeding up the evaluation
            by reducing number of queries to broker.
        :returns: Pairs in form ``(command_factory, session_proxy)``.
        :rtype: generator
        :raises:
            * :py:class:`~lmi.scripts.common.errors.LmiUnsatisfiedDependencies`
                if no condition is satisfied for at least one host. Note that
                this exception is raised at the end of evaluation. This lets
                you choose whether you want to process satisfied hosts - by
                processing the generator at once. Or whether you want to be
                sure it is satisfied by all of them - you turn the generator
                into a list.
            * :py:class:`~lmi.scripts.common.errors.LmiNoConnections`
                if no successful connection was done.
        """
        if cache is None:
            cache = dict()
        conds, default = self.get_conditionals()
        def get_cmd_factory(cmd):
            if isinstance(cmd, basestring):
                i = cmd.rindex('.')
                module = __import__(cmd[:i], fromlist=cmd[i+1:])
                return getattr(module, cmd[i+1:])
            else:
                return cmd

        session = self.session
        unsatisfied = set(session.hostnames)

        for expr, cmd in conds:
            hosts = self.eval_expr(expr, unsatisfied, cache)
            if hosts:
                yield get_cmd_factory(cmd), SessionProxy(session, hosts)
            hosts = set(hosts).union(set(session.get_unconnected()))
            unsatisfied.difference_update(hosts)
            if not unsatisfied:
                break
        if default is not None and unsatisfied:
            yield get_cmd_factory(default), SessionProxy(session, unsatisfied)
            unsatisfied.clear()
        if len(unsatisfied):
            raise errors.LmiUnsatisfiedDependencies(unsatisfied)
        if len(session) == len(session.get_unconnected()):
            raise errors.LmiNoConnections("No successful connection!")

    def get_usage(self, proper=False):
        """
        Try to get usage of any command satisfying some expression.

        :raises: Same exceptions as :py:meth:`select_cmds`.
        """
        for cmd_cls, _ in self.select_cmds():
            cmd = cmd_cls(self.app, self.cmd_name, self.parent)
            return cmd.get_usage(proper)

    def run(self, args):
        """
        Iterate over command factories with associated sessions and
        execute them with unchanged *args*.
        """
        result = 0
        for cmd_cls, session in self.select_cmds():
            cmd = cmd_cls(self.app, self.cmd_name, self.parent)
            cmd.set_session_proxy(session)
            ret = cmd.run(args)
            if result == 0:
                result = ret
        return result




# Copyright (C) 2013-2014 Michal Minar <miminar@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
"""
Module with convenient function for defining user commands.
"""

from lmi.scripts.common.command import LmiLister
from lmi.scripts.common.command import LmiCommandMultiplexer
from lmi.scripts.common.command import LmiSelectCommand
from lmi.scripts.common.command import util

def make_list_command(func,
        name=None,
        columns=None,
        verify_func=None,
        transform_func=None):
    """
    Create a command subclassed from :py:class:`~.lister.LmiLister`. Please
    refer to this class for detailed usage.

    :param func: Contents of ``CALLABLE`` property.
    :type func: string or callable
    :param string name: Optional name of resulting class. If not given,
        it will be made from the name of associated function.
    :param tuple columns: Contents of ``COLUMNS`` property.
    :param callable verify_func: Callable overriding
        py:meth:`~.endpoint.LmiEndPointCommand.verify_options` method.
    :param callable transform_func: Callable overriding
        :py:meth:`~.endpoint.LmiEndPointCommand.transform_options` method.
    :returns:  Subclass of :py:class:`~.lister.LmiLister`.
    :rtype: type
    """
    if name is None:
        if isinstance(func, basestring):
            name = func.split('.')[-1]
        else:
            name = func.__name__
        if not name.startswith('_'):
            name = '_' + name.capitalize()
    props = { 'COLUMNS' : columns
            , 'CALLABLE' : func
            , '__module__' : util.get_module_name() }
    if verify_func:
        props['verify_options'] = verify_func
    if transform_func:
        props['transform_options'] = transform_func
    return LmiLister.__metaclass__(name, (LmiLister, ), props)

def register_subcommands(command_name, usage, command_map,
        fallback_command=None):
    """
    Create a multiplexer command (a node in a tree of commands).

    :param string command_name: Name of created command. The same as will
        be given on a command line.
    :param string usage: Usage string parseable by ``docopt``.
    :param dictionary command_map: Dictionary of subcommands. Associates
        command names to their factories. It's assigned to ``COMMANDS``
        property.
    :param fallback_command: Command factory used when no command is given
        on command line.
    :type fallback_command: :py:class:`~.endpoint.LmiEndPointCommand`
    :returns: Subclass of :py:class:`~.multiplexer.LmiCommandMultiplexer`.
    :rtype: type
    """
    props = { 'COMMANDS'         : command_map
            , 'OWN_USAGE'        : True
            , '__doc__'          : usage
            , '__module__'       : util.get_module_name()
            , 'FALLBACK_COMMAND' : fallback_command }
    return LmiCommandMultiplexer.__metaclass__(command_name,
            (LmiCommandMultiplexer, ), props)

def select_command(command_name, *args, **kwargs):
    """
    Create command selector that loads command whose requirements are met.

    Example of invocation: ::

        Hardware = select_command('Hardware',
            ("Openlmi-Hardware >= 0.4.2", "lmi.scripts.hardware.current.Cmd"),
            ("Openlmi-Hardware < 0.4.2" , "lmi.scripts.hardware.pre042.Cmd"),
            default=HwMissing
        )

    Above example checks remote broker for OpenLMI-Hardware provider. If it is
    installed and its version is equal or higher than 0.4.2, command from
    ``current`` module will be used. For older registered versions command
    contained in ``pre042`` module will be loaded. If hardware provider is not
    available, HwMissing command will be loaded instead.

    .. seealso::
        Check out the grammer describing language used in these conditions at
        :py:mod:`lmi.scripts.common.versioncheck.parser`.

    :param args: List of pairs ``(condition, command)`` that are inspected in
        given order until single condition is satisfied. Associated command is
        then loaded. Command is either a reference to command class or path to
        it given as string. In latter case last dot divides module's import
        path and command name.
    :param default: This command will be loaded when no condition from *args*
        is satisfied.
    """
    props = { 'SELECT'         : args
            , 'DEFAULT'        : kwargs.get('default', None)
            , '__module__'     : util.get_module_name()
            }
    return LmiSelectCommand.__metaclass__(command_name,
            (LmiSelectCommand, ), props)




# Copyright (C) 2013-2014 Michal Minar <miminar@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
"""
This subpackage defines base classes and utility functions for declaring
commands. These serve as wrappers for functions in libraries specific to
particular provider.

Tree of these commands build a command line interface for this library.
"""

from lmi.scripts.common.command.base import LmiBaseCommand
from lmi.scripts.common.command.checkresult import LmiCheckResult
from lmi.scripts.common.command.endpoint import LmiEndPointCommand
from lmi.scripts.common.command.lister import LmiInstanceLister
from lmi.scripts.common.command.lister import LmiLister
from lmi.scripts.common.command.multiplexer import LmiCommandMultiplexer
from lmi.scripts.common.command.session import LmiSessionCommand
from lmi.scripts.common.command.select import LmiSelectCommand
from lmi.scripts.common.command.show import LmiShowInstance

from lmi.scripts.common.command.helper import make_list_command
from lmi.scripts.common.command.helper import register_subcommands
from lmi.scripts.common.command.helper import select_command



# Copyright (C) 2012-2014 Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import sys
import time
import random
import string

from lmi.shell.compat import *

from lmi.shell.LMIExceptions import LMIHandlerNamePatternError


class LMIIndicationListener(wbem.CIMIndicationListener):
    """
    Class representing indication listener, which provides a unified API for
    the server startup and shutdown and for registering an indication handler.

    :param str hostname: bind hostname
    :param int port: TCP port, where the server should listen for incoming
        messages
    :param str certfile: path to certificate file, if SSL used
    :param str keyfile: path to key file, if SSL used
    :param str trust_store: path to trust store
    """
    # Minimum replacable "X" characters in handler pattern string.
    HANDLER_MINIMUM_REPL_CHARS_COUNT = 8

    def __init__(self, hostname, port, certfile=None, keyfile=None,
                 trust_store=None):
        super(LMIIndicationListener, self).__init__(
            hostname, port, certfile, keyfile, trust_store)

    def __create_handler_name(self, handler_name_pattern):
        """
        Returns unique handler name by replacing "**X**" characters for random
        characters at the end of the handler_name_pattern.

        :param string handler_name_pattern: string containing replaceable
            characters at the end
        :returns: unique handler name
        :rtype: string
        """
        placeholder_char = "X"
        allowed_chars = string.ascii_uppercase + string.digits
        min_strength = LMIIndicationListener.HANDLER_MINIMUM_REPL_CHARS_COUNT

        def draw_string_of(strength):
            return "".join(random.choice(allowed_chars)
                           for x in xrange(strength))

        prefix = handler_name_pattern.rstrip(placeholder_char)
        strength = len(handler_name_pattern) - len(prefix)

        if strength == 0:
            return handler_name_pattern
        elif strength < min_strength:
            raise LMIHandlerNamePatternError(
                "Not enough replaceable characters provided")
        else:   # construct the handler name
            while True:
                handler_name = prefix + draw_string_of(strength)
                if handler_name not in self.handlers:
                    return handler_name

    def add_handler(self, handler_name_pattern, handler, *args, **kwargs):
        """
        Registers a handler into the indication listener. Returns a string,
        which is used for the indication recognition, when a message arrives.

        :param string handler_name_pattern: string, which may contain set of
            "**X**" characters at the end of the string. The "**X**" characters
            will be replaced by random characters and the whole string will
            form a unique string.
        :param handler: callable, which will be executed, when a indication is
            received
        :param tuple args: positional arguments for the handler
        :param dictionary kwargs: keyword arguments for the handler
        :returns: handler unique name
        :rtype: string
        """
        handler_name = self.__create_handler_name(handler_name_pattern)
        wbem.CIMIndicationListener.add_handler(
            self, handler_name, handler, *args, **kwargs)
        return handler_name



# Copyright (C) 2012-2014 Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import os
import abc
import sys
import tempfile
import subprocess
from textwrap import TextWrapper

from lmi.shell.LMIUtil import lmi_raise_or_dump_exception

from lmi.shell.LMIExceptions import LMINoPagerError


def _is_executable(fpath):
    """
    :param string fpath: path of executable
    :returns: True, if provided path is executable, False otherwise
    """
    return os.path.exists(fpath) and os.access(fpath, os.X_OK)


def _which(program):
    """
    :param string program: executable name
    :returns: full path of a selected executable
    :rtype: string
    """
    fpath, fname = os.path.split(program)
    if fpath:
        if _is_executable(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            exe_file = os.path.join(path, program)
            if _is_executable(exe_file):
                return exe_file
    return None


def _get_pager_with_params():
    """
    :returns: list containing a executable with its CLI arguments
    """
    if "PAGER" in os.environ:
        path = _which(os.environ["PAGER"])
        if path:
            return [path]
    for p in ["less", "more"]:
        path = _which(p)
        if not path:
            continue
        elif p == "less":
            return [path, "-S"]
        return [path]
    lmi_raise_or_dump_exception(LMINoPagerError("No default pager found"))
    return []


class LMITextFormatter(object):
    """
    Text formatter class. Used when printing a block of text to output stream.

    :param string text: text to be formatted
    """
    SUB_INDENT = 4

    def __init__(self, text):
        self._text = text

    def format(self, indent=0, width=80, f=sys.stdout,
               separator=True):
        """
        Formats a block of text and prints it to the output stream.

        :param int indent: number of spaces to indent the text block
        :param int width: total text block width
        :param f: output stream
        :param dictionary kwargs: supported keyword arguments
        :param bool separator: if True, there will be a new line appended after
            the formatted text; default value is True
        """
        if indent > width:
            return  # NOTE: this is wrong!
        wrapper = TextWrapper()
        wrapper.width = width - indent
        wrapper.subsequent_indent = " " * LMITextFormatter.SUB_INDENT
        for l in wrapper.wrap(self._text):
            f.write("%s%s\n" % (" " * indent, l))
        if separator:
            f.write("\n")


class LMIFormatter(object):
    """
    Abstract class for derived subclasses.
    """
    __metaclass__ = abc.ABCMeta

    def __init__(self):
        stty_dim = os.popen("stty size 2> /dev/null", "r").read().split()
        self._width = int(stty_dim[1]) if stty_dim else 80

    @abc.abstractmethod
    def format(self, indent=0, width=80, f=sys.stdout):
        """
        Formats a block of text and prints it to the output stream.

        :param int indent: number of spaces to indent the text block
        :param int width: total text block width
        :param f: output stream
        """
        pass

    def fancy_format(self, interactive):
        """
        Formats a block of text. If the LMIShell is running in interactive
        mode, pager will be used, otherwise the output will be written to
        standard output.

        :param bool interactive: defines, if to use pager
        """
        if interactive:
            tmpfile = tempfile.mkstemp()
            f = os.fdopen(tmpfile[0], "w")
            self.format(0, self._width, f)
            f.close()
            subprocess_params = _get_pager_with_params()
            if not subprocess_params:
                return
            subprocess_params.append(tmpfile[1])
            subprocess.call(subprocess_params)
            os.remove(tmpfile[1])
        else:
            self.format(0, 0, self._width)


class LMIMethodFormatter(LMIFormatter):
    """
    Method formatter is used to print out :py:class:`wbem.CIMMethod`
    representation.
    """
    def __init__(self, cim_method):
        super(LMIMethodFormatter, self).__init__()
        self._cim_method = cim_method

    def format_qualifier(self, qualif, indent, width, f):
        """
        Prints out a parameter of :py:class:`wbem.CIMMethod`.

        :param int indent: number of spaces to indent the text block
        :param int width: total text block width
        :param f: output stream
        """
        val = "[qualifier] %s %s" % (qualif.type, qualif.name)
        if qualif.value:
            if isinstance(qualif.value, list):
                val += ": { " + ", ".join(
                    ["'" + str(v) + "'" for v in qualif.value]) + " }"
            else:
                val += ": '%s'" % str(qualif.value)
        LMITextFormatter(val).format(indent, width, f)

    def format_parameter(self, param, indent, width, f):
        """
        Prints out a parameter of :py:class:`wbem.CIMMethod`.

        :param int indent: number of spaces to indent the text block
        :param int width: total text block width
        :param f: output stream
        """
        ref_classname = ""
        if param.type == "reference":
            ref_classname = " (%s)" % param.reference_class
        formatter = LMITextFormatter("[param] %s%s %s" % (
            param.type, ref_classname, param.name))
        formatter.format(indent, width, f)

    def format_method(self, method, indent, width, f):
        """
        Prints out a method of :py:class:`wbem.CIMClass`.

        :param int indent: number of spaces to indent the text block
        :param int width: total text block width
        :param f: output stream
        """
        has_args = "..." if method.parameters else ""
        formatter = LMITextFormatter(
            "[method] %s %s(%s)" % (method.return_type, method.name, has_args))
        formatter.format(indent, width, f)
        for q in method.qualifiers.itervalues():
            self.format_qualifier(q, indent+4, width, f)
        for p in method.parameters.itervalues():
            self.format_parameter(p, indent+4, width, f)

    def format(self, indent=0, width=80, f=sys.stdout):
        """
        Prints out :py:class`CIMMethod` object.

        :param int indent: number of spaces to indent the text block
        :param int width: total text block width
        :param f: output stream
        """
        self.format_method(self._cim_method, indent, width, f)


class LMIClassFormatter(LMIMethodFormatter):
    """
    Class formatter is used to print out :py:class:`wbem.CIMClass`
    representation.

    :param CIMClass cim_class: object to print out
    """
    def __init__(self, cim_class):
        # Call LMIFormatter ctor directly and use inly LMIMethodFormatter
        # methods for LMIClass documentation formatting.
        LMIFormatter.__init__(self)
        self._cim_class = cim_class

    def format_property(self, prop, indent, width, f):
        """
        Prints out a property of :py:class:`wbem.CIMClass`.

        :param int indent: number of spaces to indent the text block
        :param int width: total text block width
        :param f: output stream
        """
        is_array_str = " array" if prop.is_array else ""
        val = "[property%s] %s %s" % (is_array_str, prop.type, prop.name)
        LMITextFormatter(val).format(indent, width, f)
        for q in prop.qualifiers.itervalues():
            self.format_qualifier(q, indent+4, width, f)

    def format(self, indent=0, width=80, f=sys.stdout):
        """
        Formats out :py:class:`wbem.CIMClass` object.

        :param int indent: number of spaces to indent the text block
        :param int width: total text block width
        :param f: output stream
        """
        formatter = LMITextFormatter("Class: %s (%s)" % (
            self._cim_class.classname,
            self._cim_class.superclass))
        formatter.format(indent, width, f)
        for q in self._cim_class.qualifiers.itervalues():
            self.format_qualifier(q, indent+4, width, f)
        for p in self._cim_class.properties.itervalues():
            self.format_property(p, indent+4, width, f)
        for m in self._cim_class.methods.itervalues():
            self.format_method(m, indent+4, width, f)


class LMIInstanceFormatter(LMIFormatter):
    """
    Instance formatter is used to print out :py:class:`wbem.CIMInstance`
    representation.

    :param CIMInstance cim_instance: object to print out
    """
    def __init__(self, cim_instance):
        super(LMIInstanceFormatter, self).__init__()
        self._cim_instance = cim_instance

    def format_property(self, prop, indent, width, f):
        """
        Prints out a property of :py:class:`wbem.CIMInstance`.

        :param int indent: number of spaces to indent the text block
        :param int width: total text block width
        :param f: output stream
        """
        is_array_str = " array" if prop.is_array else ""
        val = "[property%s] %s %s" % (is_array_str, prop.type, prop.name)
        if prop.value:
            if isinstance(prop.value, list):
                val += " = { " + ", ".join(
                    ["'" + str(v) + "'" for v in prop.value]) + " }"
            else:
                val += " = '%s'" % str(prop.value)
        LMITextFormatter(val).format(indent, width, f)
        for q in prop.qualifiers.itervalues():
            self.format_qualifier(q, indent+4, width, f)

    def format(self, indent=0, width=80, f=sys.stdout):
        """
        Prints out :py:class`CIMInstance` object.

        :param int indent: number of spaces to indent the text block
        :param int width: total text block width
        :param f: output stream
        """
        formatter = LMITextFormatter(
            "Instance of %s" % self._cim_instance.classname)
        formatter.format(indent, width, f)
        for p in self._cim_instance.properties.itervalues():
            self.format_property(p, indent+4, width, f)


class LMIMofFormatter(LMIFormatter):
    """
    MOF formatter is used to print out MOF representation of a CIM object.

    :param obj: object, which has :py:meth:`tomof` method
    """
    def __init__(self, obj):
        super(LMIMofFormatter, self).__init__()
        if not hasattr(obj, "tomof") or not callable(obj.tomof):
            raise TypeError("object needs provide tomof() method")
        self._obj = obj

    def format(self, indent=0, width=80, f=sys.stdout):
        """
        Formats a MOF object and prints it to the output stream.

        :param int indent: number of spaces to indent the text block
        :param int width: total text block width
        :param f: output stream
        """
        f.write(self._obj.tomof())



# Copyright (C) 2013-2014 Michal Minar <miminar@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
"""
Module with interactive application.
"""

import cmd
import docopt
import itertools
import os
import readline
import shlex

from lmi.scripts.common import errors
from lmi.scripts.common import get_logger
from lmi.scripts.common.command import LmiBaseCommand, LmiCommandMultiplexer
from lmi.scripts._metacommand import cmdutil
from lmi.scripts._metacommand import exit

LOG = get_logger(__name__)

BUILT_INS_USAGE = """
Built-ins.

Usage:
    : cd [<path>]
    : ..
    : pwd
    : (help | -h | --help)

Description:
    cd    Nest to a subcommand. Accepts a sequence of subcommands separated
          with "/". If "/" is given, top-level command becomes the active one.
          Other supported special symbols are ".." and ".".
    ..    Make parent command the active one. Same as issuing ":cd ..".
    pwd   Print current command path.
    help  Show this help.
"""

BUILT_INS = ('..', 'cd', 'pwd', 'help')

class LmiBuiltInError(errors.LmiError):
    """ General exception concerning the use of built-in commands. """
    pass

class LmiCanNotNest(LmiBuiltInError):
    """ Raised upon invalid use of *cd* built-in command. """
    pass

class Interactive(cmd.Cmd):
    """
    Launched by the main application. It enters *shell* mode. Allows to launch
    set of commands interactively on all given hosts. Session object stays the
    same for the whole life of interactive mode.

    :param top_level_cmd: Top-level command.
    :type top_level_cmd: :py:class:`.toplevel.TopLevelCommand`
    """

    def __init__(self, top_level_cmd):
        self._top_level_cmd = top_level_cmd
        cmd.Cmd.__init__(self,
                stdin=top_level_cmd.app.stdin, stdout=top_level_cmd.app.stdout)
        self._last_exit_code = exit.EXIT_CODE_SUCCESS
        self.doc_header = 'Static commands'
        self.app.active_command = top_level_cmd
        self.load_history()

    # *************************************************************************
    # Properties
    # *************************************************************************
    @property
    def app(self):
        """ :returns: Application object. """
        return self._top_level_cmd.app

    @property
    def command_manager(self):
        """ :returns: An instance of :py:class:`~.manager.CommandManager`. """
        return self.app.command_manager

    @property
    def on_top_level_node(self):
        """
        :returns: Whether the current node is a top-level one. In other words
            we're not nested in any subcommand namespace.
        :rtype: boolean
        """
        return self._top_level_cmd is self.app.active_command

    @property
    def prompt(self):
        """
        :returns: Dynamically computed shell prompt.
        :rtype: string
        """
        if self.app.stdin.isatty():
            parents_num = 0
            node = self.app.active_command
            while node.parent is not None:
                parents_num += 1
                node = node.parent
            return '>'*parents_num + self.app.active_command.cmd_name + '> '
        return ''

    # *************************************************************************
    # Private methods
    # *************************************************************************
    def _change_to_node(self, path):
        """
        Handles any command path. It constructs an object of command
        corresponding to given path. Path can be absolute or relative.

        :param str path: Path to command. Looks similar to unix file path.
            Command names are separated with ``'/'``.
        :returns: Multiplexer command corresponding to path.
        :rtype: :py:class:`~lmi.scripts.common.command.multiplexer.LmiCommandMultiplexer`
        """
        cur_node = self.app.active_command
        if path.startswith('/'):
            if path.startswith('/lmi'):
                path = path[4:]
            else:
                path = path[1:]
            cur_node = self._top_level_cmd

        cmd_chain = os.path.normpath(path).split('/')
        for subcmd in cmd_chain:
            if subcmd == '..':
                cur_node = (  cur_node.parent
                           if cur_node.parent is not None else cur_node)
            elif subcmd and subcmd != '.':
                if not subcmd in cmdutil.get_subcommand_names(cur_node):
                    raise LmiCanNotNest('No such subcommand "%s".' %
                            "/".join(cmd_chain))
                cmd_cls = cmdutil.get_subcommand_factory(cur_node, subcmd)
                if not issubclass(cmd_cls, LmiCommandMultiplexer):
                    raise LmiCanNotNest('Can not nest to subcommand "%s" which'
                            " is not a multiplexer." % "/".join(cmd_chain))
                if not cmd_cls.has_own_usage():
                    raise LmiCanNotNest('Can not nest to subcommand "%s" which'
                            ' lacks any help message.' % "/".join(cmd_chain))
                cur_node = cmd_cls(self.app, subcmd, cur_node)
        self.app.active_command = cur_node
        return exit.EXIT_CODE_SUCCESS

    def _do_built_in_cmd(self, args):
        """
        Execute built-in command.

        :param list args: Command arguments including command name.
        """
        options = docopt.docopt(BUILT_INS_USAGE, args, help=False)
        if options['cd']:
            path = '.' if not options.get('<path>', None) else options['<path>']
            return self._change_to_node(path)
        elif options['..']:
            return self._change_to_node('..')
        elif options['pwd']:
            self.app.stdout.write("/"
                    + "/".join(self.app.active_command.get_cmd_name_parts(
                        all_parts=True)) + "\n")
        elif options['help'] or options['-h'] or options['--help']:
            self.app.stdout.write(BUILT_INS_USAGE[1:])
        return exit.EXIT_CODE_SUCCESS

    def _execute_line_parts(self, line_parts):
        """
        Try to execute given line. This method can throw various exceptions
        that needs to be handled by a caller, otherwise interactive mode will
        be terminated.

        :param list line_parts: Parsed command line arguments.
        :returns: Command's exit code.
        """
        if line_parts[0][0] == ':':
            if len(line_parts[0]) > 1:
                line_parts[0]  = line_parts[0][1:]
            else:
                line_parts = line_parts[1:]
            return self._do_built_in_cmd(line_parts)

        else:
            # let's try to run registered subcommand
            retval = self.run_subcommand(line_parts)
            if isinstance(retval, bool) or not isinstance(retval, (int, long)):
                retval = (    exit.EXIT_CODE_SUCCESS
                         if   bool(retval) or retval is None
                         else exit.EXIT_CODE_FAILURE)
            return retval

    # *************************************************************************
    # Public methods
    # *************************************************************************
    def complete(self, text, state):
        """
        Overrides parent's method so that registered commands can be also
        completed.
        """
        if state == 0:
            import readline
            origline = readline.get_line_buffer()
            line = origline.lstrip()
            stripped = len(origline) - len(line)
            begidx = readline.get_begidx() - stripped
            endidx = readline.get_endidx() - stripped
            command, _, _ = self.parseline(line)
            compfunc = self.completedefault
            if command and hasattr(self, 'complete_' + command):
                compfunc = getattr(self, 'complete_' + command)
            self.completion_matches = compfunc(text, line, begidx, endidx)
        try:
            return self.completion_matches[state]
        except IndexError:
            return None

    def completedefault(self, text, line, *_args, **_kwargs):
        """
        Tab-completion for commands known to the command manager and subcommands
        in current command namespace. Does not handle command options.
        """
        if line.startswith(':'):
            commands = BUILT_INS
        else:
            commands = set(self.completenames(text))
            commands.update(set(cmdutil.get_subcommand_names(self.app.active_command)))
        completions = sorted(n for n in commands
                if not text or n.startswith(text))
        return completions

    def clear_history(self):
        """ Clear readline history. """
        readline.clear_history()

    def default(self, line):
        """
        This is run, when line contains unknown command to ``cmd.Cmd``. It
        expects us to handle it if we know it, or print an error.

        :param str line: Line given to our shell.
        """
        try:
            line_parts = shlex.split(line)
        except ValueError as err:
            LOG().error(str(err))
            return exit.EXIT_CODE_INVALID_SYNTAX

        try:
            self._execute_line_parts(line_parts)

        except docopt.DocoptExit as err:
            # command found, but options given to it do not comply with its
            # usage string
            if '--help' in line_parts:
                return self.do_help(" ".join(
                    line_parts[:line_parts.index('--help')]))
            LOG().warn("Wrong options given: %s", line.strip())
            self.stdout.write(str(err))
            if (   line_parts[0] in cmdutil.get_subcommand_names(
                    self.app.active_command)
               and cmdutil.get_subcommand_factory(self.app.active_command,
                    line_parts[0]).is_end_point()):
                self.stdout.write("\n\nTo see a full usage string, type:\n"
                        "    help %s\n" % line_parts[0])
            else:
                self.stdout.write("\n\nTo see a list of available commands,"
                        " type:\n    help\n")
            self._last_exit_code = exit.EXIT_CODE_FAILURE

        except errors.LmiCommandNotFound as err:
            LOG().error(str(err))
            self._last_exit_code = exit.EXIT_CODE_COMMAND_NOT_FOUND

        except errors.LmiUnsatisfiedDependencies as err:
            LOG().error(str(err))
            self._last_exit_code = exit.EXIT_CODE_UNSATISFIED_DEPENDENCIES

        except errors.LmiError as err:
            LOG().error(str(err))
            self._last_exit_code = exit.EXIT_CODE_FAILURE

        except KeyboardInterrupt as err:
            LOG().debug('%s: %s', err.__class__.__name__, str(err))
            self._last_exit_code = exit.EXIT_CODE_KEYBOARD_INTERRUPT

        return self._last_exit_code

    def do_EOF(self, _arg):     #pylint: disable=C0103,R0201
        """
        Exit on End-Of-File if we are on top-level command. Otherwise change
        to parent command.
        """
        if self.app.stdin.isatty():
            self.app.stdout.write('\n')
        if self.app.stdin.isatty() and not self.on_top_level_node:
            self._change_to_node('..')
        else:
            raise errors.LmiTerminate(self._last_exit_code)

    def do_exit(self, arg):
        """
        This makes the exit command work in both (non-)interactive modes.
        """
        command = ["exit"]
        if arg:
            command.append(arg)
        return self.run_subcommand(command)

    def do_help(self, arg):
        """ Handle help subcommand. """
        if arg:
            try:
                arg_parts = shlex.split(arg)
            except ValueError as err:
                LOG().error(str(err))
                return exit.EXIT_CODE_INVALID_SYNTAX

            method_name = '_'.join(
                itertools.chain(['do'],
                    itertools.takewhile(lambda x: not x.startswith('-'),
                                        arg_parts)))
            if hasattr(self, method_name):
                return cmd.Cmd.do_help(self, arg)
        else:
            arg_parts = []

        if (   self.on_top_level_node
           and (  not arg
               or (   len(arg_parts) == 1
                  and arg not in self.app.command_manager.command_names))):
            if not self.completenames(arg):
                LOG().error(str(errors.LmiCommandNotFound(arg)))
                cmd.Cmd.do_help(self, '')
            else:
                cmd.Cmd.do_help(self, arg)
            cmd_names = set(self.command_manager)
            cmd_names.difference_update(set(self.completenames('')))
            self.print_topics(
                    "Application commands (type help <topic>):",
                    sorted(cmd_names), 15, 80)
            self.print_topics(
                    "Built-in commands (type :help):",
                    [':'+bi for bi in BUILT_INS], 15, 80)
            return exit.EXIT_CODE_SUCCESS

        return self.run_subcommand(['help'] + arg_parts)

    def emptyline(self):   #pylint: disable=R0201
        """ Do nothing for empty line. """
        pass

    def help_exit(self):
        """ Provide help for exit command. """
        cur_node = self.app.active_command
        # temporarily change to top level command where help cmd is registered
        self.app.active_command = self._top_level_cmd
        try:
            return self.run_subcommand(["help", "exit"])
        finally:
            self.app.active_command = cur_node

    def help_help(self):
        """
        Use the command manager to get instructions for "help".
        """
        cur_node = self.app.active_command
        # temporarily change to top level command where help cmd is registered
        self.app.active_command = self._top_level_cmd
        try:
            return self.run_subcommand(["help", "help"])
        finally:
            self.app.active_command = cur_node

    def load_history(self):
        """
        Load a readline history file.
        """
        if      (   self.app.config.history_max_length == 0
                or not os.path.exists(self.app.config.history_file)):
            return
        LOG().debug('Reading history file "%s"', self.app.config.history_file)
        try:
            readline.read_history_file(self.app.config.history_file)
            if      (   self.app.config.history_max_length > 0
                    and   readline.get_current_history_length()
                        > self.app.config.history_max_length):
                readline.set_history_length(self.app.config.history_max_length)
                readline.write_history_file(self.app.config.history_file)
                readline.read_history_file(self.app.config.history_file)
        except (IOError, OSError) as err:
            LOG().warn('Failed to read history file "%s".',
                    self.app.config.history_file, exc_info=err)

    def postcmd(self, stop, _line):
        """
        This is called after the ``do_*`` command to postprocess its result and
        decide whether to stop the shell. We want to stop only when
        :py:class:`lmi.scripts.common.errors.LmiError` is raised. This
        exception is catched upwards in call chain.

        :returns: Whether to stop the shell.
        :rtype: bool
        """
        return False

    def run_subcommand(self, args):
        """
        Run a subcommand given as a first item of ``args``. It must be one of
        commands registered in manager. Returns the return value of invoked
        command.

        :param list args: List of commands.
        """
        if not isinstance(args, (list, tuple)):
            raise TypeError("args must be a list")
        if len(args) < 1:
            raise ValueError("args must not be empty")
        try:
            cmd_factory = cmdutil.get_subcommand_factory(
                    self.app.active_command, args[0])
            parent = self.app.active_command
        except errors.LmiCommandNotFound:
            # When nested into a subcommand, let it handle the args if known.
            # If not known, try one of static commands.
            if not self.on_top_level_node and hasattr(self, 'do_' + args[0]):
                cmd_factory = self.app.command_manager.find_command(args[0])
                parent = self._top_level_cmd
            else:
                raise
        cmd_inst = cmd_factory(self.app, args[0], parent)
        return cmd_inst.run(args[1:])

    def save_history(self):
        """
        Saves current history of commands into the history file. If the length
        of history exceeds a maximum history file length, the history will be
        truncated.
        """
        if self.app.config.history_max_length != 0:
            LOG().debug('Writing history file "%s"',
                    self.app.config.history_file)
            if self.app.config.history_max_length > 0:
                readline.set_history_length(self.app.config.history_max_length)
            try:
                readline.write_history_file(self.app.config.history_file)
            except (IOError, OSError), err:
                LOG().warn('Failed to write history file "%s".',
                        self.app.config.history_file, exc_info=err)



# Copyright (C) 2013-2014 Michal Minar <miminar@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
"""
Subpackage containing functionality of lmi meta-command.
"""

import argparse
import logging
import sys

from lmi.scripts import common
from lmi.scripts.common import errors
from lmi.scripts._metacommand import util
from lmi.scripts._metacommand import exit
from lmi.scripts._metacommand.help import Help
from lmi.scripts._metacommand.manager import CommandManager
from lmi.scripts._metacommand.interactive import Interactive
from lmi.scripts._metacommand.toplevel import TopLevelCommand
from lmi.scripts.common.command import LmiCommandMultiplexer, LmiBaseCommand
from lmi.scripts.common.configuration import Configuration
from lmi.scripts.common.session import Session
from lmi.shell import LMIUtil

LOG = common.get_logger(__name__)

# write errors to stderr until logging is configured
logging.getLogger('').addHandler(logging.StreamHandler())

class MetaCommand(object):
    """
    Main application class. It instantiates configuration object, logging and
    then it passes control to commands.

    Example usage:

        MetaCommand().run()
    """

    def __init__(self):
        # allow exceptions in lmi shell
        LMIUtil.lmi_set_use_exceptions(True)
        # instance of CommandManager, created when first needed
        self._command_manager = None
        self.stdout = sys.stdout
        self.stderr = sys.stderr
        self.stdin = sys.stdin
        # instance of Session, created when needed
        self._session = None
        # instance of Configuration, created in setup()
        self.config = None
        # dictionary of not yet processed options, it's created in setup()
        self._options = None
        self._active_command = None

    def _configure_logging(self):
        """
        Setup logging. It expects Configuration object to be already
        initialized.

        Logging can be tuned in various ways:

            * In configuration file with options:
                * [Main] Verbosity
                * [Log] OutputFile
                * [Log] FileFormat
                * [Log] ConsoleFormat
                * [Log] ConsoleInfoFormat
                * [Log] LogToConsole
            * With command line options:
                ``-v`` flags :
                    Each such flag increases logging level of what is logged
                    into console. This overrides `[Main] Verbosity` option.
                ``-q`` :
                    Causes supression of any output made to stdout except for
                    error messages. This overrides ``[Main] Verbosity``.
                    option and ``-v`` flags.
                ``--log-file`` :
                    Output file for logging messages. This overrides ``[Log]
                    OutputFile`` option.

        Implicitly only warnings and errors are logged to the standard error
        stream without any tracebacks.
        """
        util.setup_logging(self.config, self.stderr)

    @property
    def command_manager(self):
        """
        Return instance of ``CommandManager``. It's initialized when first
        needed.

        :rtype: (``CommandManager``)
        """
        if self._command_manager is None:
            self._command_manager = CommandManager()
            self._command_manager.add_command('help', Help)
        return self._command_manager

    @property
    def session(self):
        """
        Return instance of Session. Instantiated when first needed.

        :rtype: (``Session``)
        """
        if self._session is None:
            if (   not self._options['--host']
               and not self._options['--hosts-file']):
                self._options['--host'] = [util.get_default_hostname()]
                LOG().info('No hosts given, using "%s".',
                        self._options['--host'][0])
            hostnames = []
            # credentials loaded from file
            credentials = {}
            def add_hosts(hosts, creds):
                """ Update hostnames and credentials for new data. """
                hostnames.extend(hosts)
                credentials.update(creds)
            if self._options['--hosts-file']:
                hosts_path = self._options['--hosts-file']
                try:
                    with open(hosts_path, 'r') as hosts_file:
                        add_hosts(*util.parse_hosts_file(hosts_file))
                except (OSError, IOError) as err:
                    LOG().critical('Could not read hosts file "%s": %s',
                            hosts_path, err)
                    sys.exit(1)
            add_hosts(*util.get_hosts_credentials(self._options['--host']))
            if self._options['--user']:
                credentials.update(
                        # credentials in file has precedence over --user option
                        dict((h, credentials.get(h, (self._options['--user'], '')))
                    for h in hostnames if h not in credentials
                ))
            self._session = Session(self, hostnames, credentials,
                    same_credentials=self._options['--same-credentials'])
        return self._session

    @property
    def active_command(self):
        return self._active_command
    @active_command.setter
    def active_command(self, cmd):
        if not isinstance(cmd, LmiBaseCommand):
            raise TypeError("cmd must be an instance of LmiBaseCommand")
        self._active_command = cmd

    def print_version(self):
        """ Print version of this egg to stdout. """
        self.stdout.write("%s\n" % util.get_version())

    def setup(self, options):
        """
        Initialise global Configuration object and set up logging.

        :param options: (``dict``) Dictionary of options parsed from command
            line by docopt.
        """
        conf_kwargs = {}
        if options['--config-file']:
            conf_kwargs['user_config_file_path'] = options.pop('--config-file')
        self.config = Configuration.get_instance(**conf_kwargs)
        # two mutually exclusive options
        if options['--trace'] or options['--notrace']:
            self.config.trace = bool(options['--trace'])
        if options.pop('--quiet', False):
            self.config.verbosity = Configuration.OUTPUT_SILENT
        elif options['-v'] and options['-v'] > 0:
            self.config.verbosity = options['-v']
        if options.pop('--noverify', False):
            self.config.verify_server_cert = False
        self.config.log_file = options.pop('--log-file', None)
        self._configure_logging()
        del options['--trace']
        del options['--notrace']
        del options['-v']
        self.config.namespace = options.pop('--namespace', None)
        self.config.human_friendly = options.pop('--human-friendly', None)
        self.config.no_headings = options.pop('--no-headings', None)
        self.config.lister_format = options.pop('--lister-format', None)
        # unhandled options may be used later (for session creation),
        # so let's save them
        self._options = options

    def run(self, argv):
        """
        Equivalent to the main program for the application.

        :param argv: (``list``) Input arguments and options.
            Contains all arguments but the application name.
        """
        retval = exit.EXIT_CODE_FAILURE
        cmd = TopLevelCommand(self)
        try:
            retval = cmd.run(argv)
        except Exception as exc:
            if isinstance(exc, errors.LmiUnsatisfiedDependencies):
                retval = exit.EXIT_CODE_UNSATISFIED_DEPENDENCIES
            LOG().exception(str(exc))
        if isinstance(retval, bool) or not isinstance(retval, (int, long)):
            return (    exit.EXIT_CODE_SUCCESS if bool(retval) or retval is None
                   else exit.EXIT_CODE_FAILURE)

        return retval

def main(argv=sys.argv[1:]):
    """
    Main entry point function. It just passes arguments to instantiated
    ``MetaCommand``.
    """
    return MetaCommand().run(argv)

